import matplotlib.pyplot as plt
import numpy as np
from pymatgen import Structure
from monty.serialization import loadfn

pyxtal_data = loadfn("Si_train.json")
#ucsd_train_data = loadfn("pyxtal_ff/datasets/Si/training.json")
#ucsd_test_data = loadfn("pyxtal_ff/datasets/Si/test.json")

pyxtal_energy = []
pyxtal_volume = []
#pyxtal_force = []
for i, d in enumerate(pyxtal_data):
    lat = d['lattice']
    species = d['elements']
    coords = d['coords']
    struc = Structure(lat, species, coords)
    pyxtal_energy.append(d['energy']/len(d['elements']))
    pyxtal_volume.append(struc.volume/len(d['elements']))

#u_train_energy = []
#u_train_volume = []
#for i, d in enumerate(ucsd_train_data):
#    u_train_energy.append(d['outputs']['energy']/len(d['structure']))
#    u_train_volume.append(d['structure'].volume/len(d['structure']))

#u_test_energy = []
#u_test_volume = []
#for i, d in enumerate(ucsd_test_data):
#    u_test_energy.append(d['outputs']['energy']/len(d['structure']))
#    u_test_volume.append(d['structure'].volume/len(d['structure']))


#plt.scatter(u_test_volume, u_test_energy, s=15, label='UCSD_Test', color='red')
plt.scatter(pyxtal_volume, pyxtal_energy, s=7, label='PyXtal', color='k')
#plt.scatter(u_train_volume, u_train_energy, s=7, color='deepskyblue', label='UCSD_Train')
plt.xlabel('Volume/atom (A\N{SUPERSCRIPT THREE}/atom)')
plt.ylabel('Energy/atom (eV/atom)')
plt.legend(loc='lower right')
plt.show()



from pyxtal_ff import PyXtal_FF
from pyxtal_ff.calculator import PyXtalFFCalculator, optimize, elastic_properties
from pyxtal_ff.calculator.elasticity import fit_elastic_constants
from pyxtal_ff.calculator.phonon import Phonon
from ase.optimize import BFGS
from ase import units
import numpy as np

if  __name__ == "__main__":

    from optparse import OptionParser
    from ase.build import bulk

    parser = OptionParser()
    parser.add_option("-f", "--file", dest="file",
                      help="pretrained file from pyxtal_ff, REQUIRED",
                      metavar="file")

    (options, args) = parser.parse_args()

    # load calculator
    #calc = PyXtalFFCalculator(mliap=options.file, logo=False)

    ff = PyXtal_FF(model={'system': ["Si"]}, logo=False)
    ff.run(mode='predict', mliap=options.file)
    calc = PyXtalFFCalculator(ff=ff)


    # initial structure and calculator
    si = bulk('Si', 'diamond', a=5.0, cubic=True)
    si.set_calculator(calc)

    # geometry optimization
    si = optimize(si, box=True)
    print('equlirum cell para: ', si.get_cell()[0][0])
    print('equlirum energy: ', si.get_potential_energy())
    print('equlirum stress', -si.get_stress())

    #Elastic Properties
    C, C_err = fit_elastic_constants(si, symmetry='cubic', optimizer=BFGS)
    C /= units.GPa

    print("Bulk modulus, Shear modulus, Young's modulus, Poisson's ratio")
    k1, g1, e1, v1, k2, g2, e2, v2, k3, g3, e3, v3 = elastic_properties(C)
    print(k1, g1, e1, v1)
    print(k2, g2, e2, v2)
    print(k3, g3, e3, v3)

    # Phonon properties
    supercell_matrix=np.diag([2,2,2])
    ph = Phonon(si, calc, supercell_matrix=supercell_matrix) 
    force_constants = ph.get_force_constants()
    ph.phonon.set_force_constants(force_constants)

    ph.phonon.auto_band_structure()
    #path = [[[0, 0, 0], [0.5, 0, 0.5], [0.375, 0.375, 0.75], [0, 0, 0], [0.5, 0.5, 0.5]]]
    #labels = ["$\\Gamma$", "X", "K", "$\\Gamma$", "L"]
    #qpoints, connections = get_band_qpoints_and_path_connections(path, npoints=51)
    #ph.phonon.run_band_structure(qpoints, path_connections=connections, labels=labels)

    ph.phonon.set_mesh([10,10,10])
    ph.phonon.set_total_DOS()
    ph.phonon.plot_band_structure_and_dos().savefig('Si_phonon.png', dpi=300)
    print('The detected space group is {:s} with a tolerance of {:.2e}'.format(
          ph.phonon._symmetry.dataset['international'], ph.phonon._symprec))



import numpy as np
from ase import units
from ase.optimize import LBFGS
from ase.optimize.fire import FIRE
from ase.constraints import ExpCellFilter
from ase.calculators.calculator import Calculator, all_changes
from ase.spacegroup.symmetrize import FixSymmetry, check_symmetry
from pyxtal_ff import PyXtal_FF
from pyxtal_ff.utilities import compute_descriptor
from pyxtal_ff.utilities.base_potential import ZBL
np.set_printoptions(formatter={'float': '{: 8.4f}'.format})


class PyXtalFFCalculator(Calculator):
    implemented_properties = ['energy', 'forces', 'stress']
    nolabel = True

    def __init__(self, style='ase', **kwargs):
        self.style = style
        Calculator.__init__(self, **kwargs)

    def calculate(self, atoms=None,
                  properties=['energy'],
                  system_changes=all_changes):

        Calculator.calculate(self, atoms, properties, system_changes)

        #chem_symbols = list(set(atoms.get_chemical_symbols()))
        #self.ff = PyXtal_FF(model={'system': chem_symbols}, logo=self.parameters.logo)
        #self.ff.run(mode='predict', mliap=self.parameters.mliap)

        # base potential
        if self.parameters.ff._descriptors['base_potential']:
            self.base_potential = ZBL(self.parameters.ff._descriptors['base_potential']['inner'],
                                      self.parameters.ff._descriptors['base_potential']['outer'],
                                      atomic_energy=True)
            base_results = self.base_potential.calculate(atoms)
            base_energy = base_results['energy']
            base_forces = base_results['force']
            base_stress = base_results['stress'] # eV/A^3
            base_energies = base_results['energies']
        else:
            base_energy, base_forces, base_stress = 0., 0., 0.
            base_energies = 0.

        desp = compute_descriptor(self.parameters.ff._descriptors, atoms)
        energies, forces, stress = self.parameters.ff.model.calculate_properties(desp, bforce=True, bstress=True)

        self.desp = desp
        self.results['energies'] = energies + base_energies
        self.results['energy'] = energies.sum() + base_energy
        self.results['free_energy'] = energies.sum() + base_energy
        self.results['forces'] = forces + base_forces

        # pyxtal_ff and lammps uses: xx, yy, zz, xy, xz, yz
        # ase uses: xx, yy, zz, yz, xz, xy
        # from GPa to eV/A^3
        self.results['stress_zbl'] = base_stress/units.GPa
        self.results['energy_zbl'] = base_energy
        self.results['forces_zbl'] = base_forces
        self.results['stress_ml'] = stress 
        self.results['energy_ml'] = energies.sum()
        self.results['forces_ml'] = forces


        # ase counts the stress differently
        if self.style == 'ase':
            self.results['stress'] = -(stress * units.GPa + base_stress)[[0, 1, 2, 5, 4, 3]]
        else:
            self.results['stress'] = self.results['stress_zbl'] + self.results['stress_ml']

    def __str__(self):
        s = "\nASE calculator with pyxtal_ff force field\n"
        return s

    def __repr__(self):
        return str(self)

    def print_stresses(self):
        print("stress_ml (GPa, xx, yy, zz, xy, xz, yz):", self.results["stress_ml"])
        print("stress_zbl(GPa, xx, yy, zz, xy, xz, yz):", self.results['stress_zbl'])

    def print_energy(self):
        print("energy_ml (eV):", self.results["energy_ml"])
        print("energy_zbl(eV):", self.results['energy_zbl'])

    def print_forces(self):
        print("forces (eV/A)")
        for f1, f2 in zip(self.results["forces_ml"], self.results['forces_zbl']):
            print("{:8.3f} {:8.3f} {:8.3f} -> {:8.3f} {:8.3f} {:8.3f}".format(*f1, *f2))

    def print_all(self):
        self.print_energy()
        self.print_forces()
        self.print_stresses()


def elastic_properties(C):
    Kv = C[:3,:3].mean()
    Gv = (C[0,0]+C[1,1]+C[2,2] - (C[0,1]+C[1,2]+C[2,0]) + 3*(C[3,3]+C[4,4]+C[5,5]))/15
    Ev = 1/((1/(3*Gv))+(1/(9*Kv)))
    vv  = 0.5*(1-((3*Gv)/(3*Kv+Gv))); 

    S = np.linalg.inv(C)
    Kr = 1/((S[0,0]+S[1,1]+S[2,2])+2*(S[0,1]+S[1,2]+S[2,0])) 
    Gr = 15/(4*(S[0,0]+S[1,1]+S[2,2])-4*(S[0,1]+S[1,2]+S[2,0])+3*(S[3,3]+S[4,4]+S[5,5])) 
    Er = 1/((1/(3*Gr))+(1/(9*Kr))) 
    vr = 0.5*(1-((3*Gr)/(3*Kr+Gr))) 

    Kh = (Kv+Kr)/2    
    Gh = (Gv+Gr)/2    
    Eh = (Ev+Er)/2    
    vh = (vv+vr)/2   
    return Kv, Gv, Ev, vv, Kr, Gr, Er, vr, Kh, Gh, Eh, vh

def optimize(atoms, sym=True, box=False, method='FIRE', fmax=0.01, steps=1000, logfile='ase.log'):
    if sym:
        atoms.set_constraint(FixSymmetry(atoms))
    if box:
        ecf = ExpCellFilter(atoms)
        if method == 'FIRE':
            dyn = FIRE(ecf, logfile=logfile)
        else:
            dyn = LBFGS(ecf, logfile=logfile)
    else:
        if method == 'FIRE':
            dyn = FIRE(atoms, logfile=logfile)
        else:
            dyn = FIRE(atoms, logfile=logfile)

    dyn.run(fmax=fmax, steps=steps)
    atoms.set_constraint()
    return atoms

if  __name__ == "__main__":

    from optparse import OptionParser
    from ase.build import bulk

    parser = OptionParser()
    parser.add_option("-f", "--file", dest="file",
                      help="pretrained file from pyxtal_ff, REQUIRED",
                      metavar="file")

    (options, args) = parser.parse_args()
    print(options.file)
    calc = PyXtalFFCalculator(mliap=options.file, logo=False)
    si = bulk('Si', 'diamond', a=5.459, cubic=True)
    si.set_calculator(calc)
    print(si.get_potential_energy())
    print(si.get_forces())
    print(si.get_stress())

    box = mushybox(si)
    dyn = BFGS(box)
    dyn.run(fmax=0.01)
    print('equlirum cell para: ', si.get_cell()[0][0])
    print('equlirum energy: ', si.get_potential_energy())

    Cijs, names, C = elastic_tensor(si, calc)
    for name, Cij in zip(names, Cijs):
        print("{:s}: {:8.2f}(GPa)".format(name, Cij))

    print(elastic_properties(C))



from copy import deepcopy

class gradient():
    """
    A class to compute the numerical gradient for different functions (df/dx)
    
    Parameters
    ----------
    fun: callable
        The objective function to be evaluated.
            `fun(x, *args) -> float`

    args: tuple
        arguments passed to the objective function

    epsilon: the stepwidth to evalute the gradients.
    **

    Usage:
    dfdx = gradient(f, args).dfdx()

    """
    def __init__(self, fun, args=(), epsilon=1e-6):
        self.fun = fun
        self.args = args
        self.epsilon = epsilon

    def get_perturbed_args(self, args, ids=[0,1], directions=['plus','plus']):
        args_list = deepcopy(list(args))
        for id, direction in zip(ids, directions):
            x = args_list[id]
            if direction == 'plus':
                x += self.epsilon
            elif direction == 'minus':
                x -= self.epsilon
        return tuple(args_list)

    def dfdx(self, id=0):
        """
        compute 2nd derivative d2fdx2
        """
        x0 = self.args[id]
        args_plus = self.get_perturbed_args(self.args, ids=[id], directions=['plus'])
        args_minus = self.get_perturbed_args(self.args, ids=[id], directions=['minus'])
        f_plus = self.fun(*tuple(args_plus))
        f_minus = self.fun(*tuple(args_minus))
        return (f_plus - f_minus)/(2*self.epsilon)

    def d2fdx2(self, id=0):
        """
        compute 2nd derivative d2fdx2
        """
        return self.d2fdxdy(ids=[id, id])

    def d2fdxdy(self, ids=[0, 1]):
        """
        compute 2nd derivative d2fdxdy
        Not working at the moment
        """
        args_plus_plus = self.get_perturbed_args(self.args, ids, directions=['plus', 'plus'])        
        args_plus_minus = self.get_perturbed_args(self.args, ids, directions=['plus', 'minus'])        
        args_minus_plus = self.get_perturbed_args(self.args, ids, directions=['minus', 'plus'])        
        args_minus_minus = self.get_perturbed_args(self.args, ids, directions=['minus', 'minus'])        
        f1 = self.fun(*tuple(args_plus_plus))
        f2 = self.fun(*tuple(args_plus_minus))
        f3 = self.fun(*tuple(args_minus_plus))
        f4 = self.fun(*tuple(args_minus_minus))
        return (f1-f2-f3+f4)/(4*self.epsilon**2)

if __name__ == '__main__':
    import numpy as np

    def f(x, y):
        return np.exp(x**2, y**2)
        #return x**2 + y**2

    x0 = np.linspace(0,1,5)
    y0 = np.linspace(0,1,5)
    grad = gradient(f, args=(x0, y0))
    print('\ninput x values:')
    print(x0)
    print('\ncomputing dfdx:')
    print(grad.dfdx())
    print('\ncomputing df2dx2:')
    print(grad.d2fdx2())
    print('\ncomputing df2dxdy:')
    print(grad.d2fdxdy([0,1]))



import torch
from torch.optim import Optimizer
from functools import reduce

from scipy.optimize import fmin_l_bfgs_b
import numpy as np
eps = np.finfo('double').eps


class LBFGSScipy(Optimizer):
    """Wrap L-BFGS algorithm, using Scipy routines.

    .. warning::
        This optimizer doesn't support per-parameter options and parameter
        groups (there can be only one).

    .. warning::
        Right now CPU only

    .. note::
        This is a very memory intensive optimizer (it requires additional
        ``param_bytes * (history_size + 1)`` bytes). If it doesn't fit in memory
        try reducing the history size, or use a different algorithm.

    Arguments:
        max_iter (int): maximal number of iterations per optimization step
            (default: 20)
        max_eval (int): maximal number of function evaluations per optimization
            step (default: max_iter * 1.25).
        tolerance_grad (float): termination tolerance on first order optimality
            (default: 1e-5).
        tolerance_change (float): termination tolerance on function
            value/parameter changes (default: 1e-9).
        history_size (int): update history size (default: 100).
    """
    def __init__(self, params, max_iter=20, max_eval=None,
                 tolerance_grad=1e-5, tolerance_change=1e-9, history_size=10):
        if max_eval is None:
            max_eval = max_iter * 5 // 4
        defaults = dict(max_iter=max_iter, max_eval=max_eval,
                        tolerance_grad=tolerance_grad, tolerance_change=tolerance_change,
                        history_size=history_size)
        super(LBFGSScipy, self).__init__(params, defaults)

        if len(self.param_groups) != 1:
            raise ValueError("LBFGS doesn't support per-parameter options "
                             "(parameter groups)")

        self._params = self.param_groups[0]['params']
        self._numel_cache = None

        self._n_iter = 0
        self._last_loss = None

    def _numel(self):
        if self._numel_cache is None:
            self._numel_cache = reduce(lambda total, p: total + p.numel(), self._params, 0)
        return self._numel_cache

    def _gather_flat_grad(self):
        views = []
        for p in self._params:
            if p.grad is None:
                view = p.data.new(p.data.numel()).zero_()
            elif p.grad.data.is_sparse:
                view = p.grad.data.to_dense().view(-1)
            else:
                view = p.grad.data.view(-1)
            views.append(view)
        return torch.cat(views, 0)

    def _gather_flat_params(self):
        views = []
        for p in self._params:
            if p.data.is_sparse:
                view = p.data.to_dense().view(-1)
            else:
                view = p.data.view(-1)
            views.append(view)
        return torch.cat(views, 0)

    def _distribute_flat_params(self, params):
        offset = 0
        for p in self._params:
            numel = p.numel()
            # view as to avoid deprecated pointwise semantics
            p.data = params[offset:offset + numel].view_as(p.data)
            offset += numel
        assert offset == self._numel()

    def step(self, closure):
        """Performs a single optimization step.

        Arguments:
            closure (callable): A closure that reevaluates the model
                and returns the loss.
        """
        assert len(self.param_groups) == 1

        group = self.param_groups[0]
        max_iter = group['max_iter']
        max_eval = group['max_eval']
        tolerance_grad = group['tolerance_grad']
        tolerance_change = group['tolerance_change']
        history_size = group['history_size']

        def wrapped_closure(flat_params):
            """closure must call zero_grad() and backward()"""
            flat_params = torch.from_numpy(flat_params)
            self._distribute_flat_params(flat_params)
            loss = closure()
            self._last_loss = loss
            loss = loss.item()
            flat_grad = self._gather_flat_grad().numpy()
            return loss, flat_grad

        def callback(flat_params):
            self._n_iter += 1
            print('\nIteration %i :' % (self._n_iter))
            #print('    Loss %12.6f:' % (self._last_loss.item()))

        initial_params = self._gather_flat_params()

        fmin_l_bfgs_b(wrapped_closure, initial_params, maxiter=max_iter,
                      maxfun=max_eval,
                      factr=tolerance_change / eps, pgtol=tolerance_grad, epsilon=0,
                      m=history_size,
                      callback=callback)



#!/usr/bin/env  python
# encoding: utf-8
import gc
import sys, os
import time
import json
import shelve
import numpy as np
from torch import save, load
np.set_printoptions(threshold=sys.maxsize)

import matplotlib as mpl
mpl.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import matplotlib.ticker as mticker
plt.style.use("ggplot")

from pyxtal_ff.utilities.elements import Element
eV2GPa = 160.21766

class PR():
    """ Atom-centered Polynomial Regression (PR) model. PR utilizes
    linear regression to predict the energy and forces based on 
    the atom-centered descriptors as the input values.
        
    Parameters
    ----------
    elements: list
         A list of atomic species in the crystal system.
    force_coefficient: float
        This parameter is used in the penalty function to scale the force
        contribution relative to the energy.
    stress_coefficient: float
        This parameter is used as the balance parameter scaling
        the stress contribution relative to the energy.
    stress_group: list of strings
        Only the intended group will be considered in stress training.
    order: int
        The order of the polynomial. Order 1 is for linear and order 2 is for 
        quadratic.
    path: str
        The path of the directory where everything is saved.
    alpha: float
        L2 penalty (regularization term) parameter.
    norm: int (*PR)
        This argument defines a model to calculate the regularization
        term. It takes only 1 or 2 as its value: Manhattan or Euclidean 
        norm, respectively. If alpha is None, norm is ignored.
    d_max: int
        The maximum number of descriptors (d) used in Linear Regression model.
    """
    def __init__(self, elements, force_coefficient, stress_coefficient, 
                 stress_group, order, path, alpha, norm, d_max=None):
        self.force_coefficient = force_coefficient
        self.stress_coefficient = stress_coefficient
        self.stress_group = stress_group
        self.elements = sorted(elements)
        self.order = order
        if order == 1:
            self.quadratic = False
        elif order == 2:
            self.quadratic = True
        else:
            raise ValueError("Order must be 1 or 2")
        self.path = path

        self.alpha = alpha
        self.norm = norm
        self.d_max = d_max
        self.unit = 'eV'


    def train(self, TrainData, optimizer):
        """ Fitting Linear Regression model. """
        db = shelve.open(self.path+TrainData)
        self.no_of_structures = len(list(db.keys()))

        # d_max is the total number of descriptors used.
        if self.d_max is None:
            self.d_max = db['0']['x'].shape[1]
        else:
            # d_max has to be less or equal than total descriptors.
            assert self.d_max <= len(db['0']['x'].shape[1]),\
                    "d_max is larger than total descriptors."

        if self.stress_coefficient and (self.stress_group is None):
            sg = []
            for i in range(self.no_of_structures):
                if db[str(i)]['group'] not in sg:
                    sg.append(db[str(i)]['group'])
            self.stress_group = sg

        db.close()
        
        print(f"Order              : {self.order}")
        if self.order == 1:
            print(f"No of parameters   : {self.d_max+1}")
        else:
            print(f"No of parameters   : {(self.d_max+1)*(self.d_max+2)//2}")
        print(f"No of structures   : {self.no_of_structures}")
        print(f"Force_coefficient  : {self.force_coefficient}")
        print(f"Stress_coefficient : {self.stress_coefficient}")
        print(f"alpha              : {self.alpha}")
        print(f"norm               : {self.norm}\n")

        t0 = time.time()
        y, w = self.parse_features(TrainData)

        X = self.parse_descriptors(TrainData,
            fc=self.force_coefficient, sc=self.stress_coefficient)
        
        self.coef_ = self.LinearRegression(X, y, w, self.alpha, self.norm)
        
        t1 = time.time()
        print("The training time: {:.2f} s".format(t1-t0))
        
    
    def evaluate(self, data, figname):
        """ Evaluating the train or test data set. """
        db = shelve.open(self.path+data)
        
        energy, force, stress = [], [], [] # true
        _energy, _force, _stress = [], [], [] # predicted
 
        for i in range(len(list(db.keys()))):
            no_of_atoms = len(db[str(i)]['force'])
            Energy, Force, Stress = self.calculate_properties(db[str(i)],       # Energy per atom
                                    self.force_coefficient, self.stress_coefficient)
            
            # Store energy into list
            true_energy = db[str(i)]['energy'] / no_of_atoms
            energy.append(true_energy)
            _energy.append(Energy.sum() / no_of_atoms)

            if self.force_coefficient:
                true_force = np.ravel(db[str(i)]['force'])
                Force = np.ravel(Force)
                for m in range(len(true_force)):
                    force.append(true_force[m])
                    _force.append(Force[m])

            if self.stress_coefficient and (db[str(i)]['group'] in self.stress_group):
                true_stress = np.array(db[str(i)]['stress'])#.flat[[0,3,5,3,1,4,5,4,2]]
                Stress = np.ravel(Stress)
                for m in range(len(true_stress)):
                    stress.append(true_stress[m])
                    _stress.append(Stress[m])

        energy, force, stress = np.asarray(energy), np.asarray(force), np.asarray(stress)
        _energy, _force, _stress = np.asarray(_energy), np.asarray(_force), np.asarray(_stress)

        # Dump the true and predicted values into text file.
        self.dump_evaluate(_energy, energy, filename=figname[:-4]+'Energy.txt')
        if self.force_coefficient:
            self.dump_evaluate(_force, force, filename=figname[:-4]+'Force.txt')
        if self.stress_coefficient:
            self.dump_evaluate(_stress, stress, filename=figname[:-4]+'Stress.txt')
        
        # Calculate the statistical metrics for energy.
        E_mae = self.mean_absolute_error(energy, _energy)
        E_mse = self.mean_squared_error(energy, _energy)
        E_r2 = self.r2_score(energy, _energy)
        print("The results for energy: ")
        print("    Energy R2     {:8.6f}".format(E_r2))
        print("    Energy MAE    {:8.6f}".format(E_mae))
        print("    Energy RMSE   {:8.6f}".format(E_mse))

        # Plotting the energy results.
        energy_str = 'Energy: r2({:.4f}), MAE({:.4f} {}/atom)'. \
                     format(E_r2, E_mae, self.unit)
        plt.title(energy_str)
        plt.scatter(energy, _energy, label='Energy', s=5)
        plt.legend(loc=2)
        plt.xlabel('True ({}/atom)'.format(self.unit))
        plt.ylabel('Prediction ({}/atom)'.format(self.unit))
        plt.tight_layout()
        plt.savefig(self.path+'Energy_'+figname)
        plt.close()
        print("The energy figure is exported to: {:s}".format(self.path+'Energy_'+figname))
        print("\n")

        if self.force_coefficient:
            F_mae = self.mean_absolute_error(force, _force)
            F_mse = self.mean_squared_error(force, _force)
            F_r2 = self.r2_score(force, _force)
            print("The results for force: ")
            print("    Force R2      {:8.6f}".format(F_r2))
            print("    Force MAE     {:8.6f}".format(F_mae))
            print("    Force RMSE    {:8.6f}".format(F_mse))

            # Plotting the forces results.
            length = 'A'
            if self.unit == 'Ha':
                length == 'Bohr'
            force_str = 'Force: r2({:.4f}), MAE({:.3f} {}/{})'. \
                        format(F_r2, F_mae, self.unit, length)
            plt.title(force_str)
            plt.scatter(force, _force, s=5, label='Force')
            plt.legend(loc=2)
            plt.xlabel('True ({}/{})'.format(self.unit, length))
            plt.ylabel('Prediction ({}/{})'.format(self.unit, length))
            plt.tight_layout()
            plt.savefig(self.path+'Force_'+figname)
            plt.close()
            print("The force figure is exported to: {:s}".format(self.path+'Force_'+figname))
            print("\n")

        else:
            F_mae, F_mse, F_r2 = None, None, None

        if self.stress_coefficient:
            S_mae = self.mean_absolute_error(stress, _stress)
            S_mse = self.mean_squared_error(stress, _stress)
            S_r2 = self.r2_score(stress, _stress)
            print("The results for stress: ")
            print("    Stress R2      {:8.6f}".format(S_r2))
            print("    Stress MAE     {:8.6f}".format(S_mae))
            print("    Stress RMSE    {:8.6f}".format(S_mse))

            # Plotting the stress results.
            stress_str = 'Stress: r2({:.4f}), MAE({:.3f} GPa)'. \
                        format(S_r2, S_mae)
            plt.title(stress_str)
            plt.scatter(stress, _stress, s=5, label='Stress')
            plt.legend(loc=2)
            plt.xlabel('True (GPa)')
            plt.ylabel('Prediction (GPa)')
            plt.tight_layout()
            plt.savefig(self.path+'Stress_'+figname)
            plt.close()
            print("The stress figure is exported to: {:s}".format(self.path+'Stress_'+figname))
            print("\n")
        else:
            S_mae, S_mse, S_r2 = None, None, None

        return (E_mae, E_mse, E_r2, F_mae, F_mse, F_r2, S_mae, S_mse, S_r2)


    def LinearRegression(self, X, y, w, alpha, norm=2):
        """ Perform linear regression. """
        
        m = X.shape[1] # The shape of the descriptors

        _X = X * np.sqrt(np.expand_dims(w, axis=1))
        _y = y * np.sqrt(w)

        if self.alpha:
            if norm == 1:
                theta = np.linalg.lstsq(_X.T.dot(_X), _X.T.dot(_y) - alpha, rcond=None)[0]
            elif norm == 2:
                theta = np.linalg.lstsq(_X.T.dot(_X) + alpha * np.identity(m), _X.T.dot(_y), rcond=None)[0]
            else:
                msg = f"Regularization with {norm} norm is not implemented yet."
                raise NotImplementedError(msg)
        else:
            theta = np.linalg.lstsq(_X, _y, rcond=None)[0]

        return theta


    def save_checkpoint(self, des_info, filename=None):
        """ Save Polynomial Regression model to PyTorch. """
        _filename = self.path

        if filename:
            _filename += filename
        else:
            _filename += 'PolyReg-checkpoint.pth'

        checkpoint = {'elements': self.elements,
                      'algorithm': 'PR',
                      'force_coefficient': self.force_coefficient,
                      'path': self.path,
                      'quadratic': self.quadratic,
                      'coef_': self.coef_,
                      'des_info': des_info}

        save(checkpoint, _filename)
        if des_info['type'] in ['SNAP', 'snap', 'SO3', 'SOAP']:
            self.save_weights_to_txt(des_info)
        print("The Linear Regression Potential is exported to {:s}".format(_filename))
        print("\n")


    def save_weights_to_txt(self, des_info):
        """ Saving the model weights to txt file. """
        with open(self.path+"PR_weights.txt", "w") as f:
            f.write("# Polynomial Regression weights generated in PyXtal_FF \n")
            f.write("# total_species ncoefficient \n\n")
            f.write(f"{len(self.elements)} {self.d_max+1} \n")
            count = 0
            for element in self.elements:
                #if des_info['type'] in ['SNAP', 'snap']:
                #    f.write(f"{element} 0.5 {des_info['weights'][element]} \n")
                #else:
                #    f.write(f"{element} \n")
                for _ in range(self.d_max+1):
                    f.write(f"{self.coef_[count]} \n")
                    count += 1

        with open(self.path+"DescriptorParam.txt", "w") as f:
            f.write("# Descriptor parameters generated in PyXtal_FF \n\n")
            f.write("# Required \n")
            f.write(f"rcutfac {des_info['Rc']} \n")
            
            if des_info['type'] in ['SO3', 'SOAP']:
                f.write(f"nmax {des_info['parameters']['nmax']} \n")
                f.write(f"lmax {des_info['parameters']['lmax']} \n")
                f.write(f"alpha {des_info['parameters']['alpha']} \n\n")
            else:
                f.write(f"twojmax {des_info['parameters']['lmax']*2} \n\n")

            f.write("# Elements \n\n")
            f.write(f"nelems {len(self.elements)} \n")
            f.write("elems ")
            for element in self.elements:
                f.write(f"{element} ")
            f.write("\n")

            if des_info['type'] in ['snap', 'SNAP', 'SO3', 'SOAP']:
                f.write("radelems ")
                for element in self.elements:
                    f.write("0.5 ")
                f.write("\n")

                if des_info['type'] in ['snap', 'SNAP']:
                    f.write("welems ")
                    for element in self.elements:
                        f.write(f"{des_info['weights'][element]} ")
                    f.write("\n\n")
                else:
                    f.write("welems ")
                    ele = Element(self.elements)
                    atomic_numbers = ele.get_Z()
                    for num in atomic_numbers:
                        f.write(f"{num} ")
                    f.write("\n\n")

            if des_info['type'] in ['snap', 'SNAP']:
                f.write(f"rfac0 {des_info['parameters']['rfac']} \n")
                f.write(f"rmin0 0 ")
                f.write("\n")
                f.write("switchflag 1 \n")
                f.write("bzeroflag 0 \n")


    def load_checkpoint(self, filename=None):
        """ Load Polynomial Regression file from PyTorch. """
        checkpoint = load(filename)

        # Inconsistent algorithm.
        if checkpoint['algorithm'] != 'PR':
            msg = "The loaded algorithm is not Polynomial Regression."
            raise NotImplementedError(msg)
        
        # Check the consistency with the system of elements
        msg = f"The system, {self.elements}, are not consistent with "\
                    +"the loaded system, {checkpoint['elements']}."

        if len(self.elements) != len(checkpoint['elements']):
            raise ValueError(msg)
        
        for i in range(len(self.elements)):
            if self.elements[i] != checkpoint['elements'][i]:
                raise ValueError(msg)
        
        self.coef_ = checkpoint['coef_']
        self.quadratic = checkpoint['quadratic']

        return checkpoint['des_info']


    def calculate_properties(self, descriptor, bforce=True, bstress=False):
        """ A routine to compute energy, forces, and stress.
        
        Parameters:
        -----------
        descriptor: list
            list of x, dxdr, and rdxdr.
        benergy, bforce, bstress: bool
            If False, excluding the property from calculation.

        Returns:
        --------
        energy: float
            The predicted energy
        forces: 2D array [N_atom, 3] (if dxdr is provided)
            The predicted forces
        stress: 2D array [3, 3] (if rdxdr is provided)
            The predicted stress
        """
        no_of_atoms = len(descriptor['elements'])
        energies, force, stress = np.zeros([no_of_atoms]), np.zeros([no_of_atoms, 3]), np.zeros([6])
        
        X = self.parse_descriptors({'0': descriptor}, fc=bforce, sc=bstress, train=False)
        
        _y = np.dot(X, self.coef_) # Calculate properties

        energies = _y[:no_of_atoms]

        #energy = _y[0] / no_of_atoms # get energy/atom
        
        if bforce: # get force
            force += np.reshape(_y[no_of_atoms:no_of_atoms+(no_of_atoms*3)], (no_of_atoms, 3))

        if bstress: # get stress
            stress += _y[-6:]*eV2GPa # in GPa
        
        return energies, force, stress


    def mean_absolute_error(self, true, predicted):
        """ Calculate mean absolute error of energy or force. """
        return sum(abs(true-predicted)/len(true))


    def mean_squared_error(self, true, predicted):
        """ Calculate mean square error of energy or force. """
        return np.sqrt(sum((true-predicted) ** 2 /len(true)))


    def r2_score(self, true, predicted):
        """ Calculate the r square of energy or force. """
        t_bar = sum(true)/len(true)
        square_error = sum((true-predicted) ** 2)
        true_variance = sum((true-t_bar) ** 2)
        return 1 - square_error / true_variance


    def dump_evaluate(self, predicted, true, filename):
        """ Dump the evaluate results to text files. """
        absolute_diff = np.abs(np.subtract(predicted, true))
        combine = np.vstack((predicted, true, absolute_diff)).T
        np.savetxt(self.path+filename, combine, header='Predicted True Diff', fmt='%.7e')


    def parse_descriptors(self, data, fc=True, sc=False, train=True):
        """ Parse descriptors and its gradient to 2-D array. 
        
        Returns
        -------
        X: 2-D array [n+m*3, d]
            d is the total number of descriptors, n is the total
            number of structures, and m is the total number atoms
            in the entire structures. If force_coefficient is None,
            X has the shape of [n, d].
        """
        if train:
            db = shelve.open(self.path+data)
            no_of_structures = self.no_of_structures
            no_of_atoms = self.no_of_atoms
            stress_components = self.stress_components
        else:
            db = data
            no_of_structures = 1 # 1 for train is false
            no_of_atoms = len(data['0']['elements']) if fc else 0
            stress_components = 6 if sc else 0
        
        # Determine the total number of descriptors based on SNAP or qSNAP.
        # Note: d_max != self.d_max
        if self.d_max is None: #enable calculator works
            self.d_max = len(db['0']['x'][0])

        if self.quadratic:
            d_max = (self.d_max**2+3*self.d_max)//2 # (d^2 + 3*d) / 2
        else:
            d_max = self.d_max

        columns = (1+d_max)*len(self.elements)
        rows = no_of_structures if train else no_of_atoms
        rows += no_of_atoms * 3 if fc else 0 # x, y, and z
        rows += stress_components if sc else 0 # xx, xy, xz, ..., zz
        
        X = np.zeros([rows, columns])

        # Fill in X.
        xcount = 0
        for i in range(no_of_structures):
            data = db[str(i)]
            if train:
                try:
                    _group = True if data['group'] in self.stress_group else False
                except:
                    _group = False
            else:
                _group = True
           
            if self.quadratic:
                _x = data['x'][:, :self.d_max]
                x = np.zeros((len(_x), d_max))
                x[:, :self.d_max] += data['x'][:, :self.d_max]
                if fc:
                    _dxdr = data['dxdr'][:, :self.d_max, :]
                    dxdr = np.zeros([_dxdr.shape[0], d_max, 3])
                    dxdr[:, :self.d_max, :] += _dxdr
                    seq = data['seq']
                
                #if train:
                if sc and _group: #(data['group'] in self.stress_group):
                    _rdxdr = data['rdxdr'][:, :self.d_max, :]
                    rdxdr = np.zeros([len(_x), d_max, 6])
                    rdxdr[:, :self.d_max, :] += _rdxdr
                
                # self-term for x, dxdr, and rdxdr
                x_square = 0.5 * _x ** 2
                if fc:
                    dxdr_square = np.zeros_like(_dxdr)
                    for i in range(len(_x)):
                        arr = np.where(seq[:, 0]==i)[0]
                        dxdr_square[arr] = np.einsum('jkl, k->jkl', _dxdr[arr], _x[i])
                    #dxdr_square = np.einsum('ijkl,ik->ijkl', _dxdr, _x)
                if sc and _group:
                    rdxdr_square = np.einsum('ijk,ij->ijk', _rdxdr, _x)
                
                dcount = self.d_max
                for d1 in range(self.d_max):
                    # Cross term for x, dxdr, and rdxdr
                    x_d1_d2 = np.einsum('i, ij->ij', _x[:, d1], _x[:, d1+1:])
                    if fc:
                        shp = _dxdr.shape
                        dxdr_d1_d2 = np.zeros([shp[0], shp[1]-1-d1, shp[2]])
                        for i in range(len(_x)):
                            arr = np.where(seq[:, 0]==i)[0]
                            dxdr_d1_d2[arr] = np.einsum('ij,k->ikj', _dxdr[arr][:, d1, :], _x[i, d1+1:]) + \
                                    (_dxdr[arr][:, d1+1:, :] * _x[i, d1])

                        #dxdr_d1_d2 = np.einsum('ijl,ik->ijkl', _dxdr[:, :, d1, :], _x[:, d1+1:]) + \
                                #             np.einsum('ijkl,i->ijkl', _dxdr[:, :, d1+1:, :], _x[:, d1])
                    if sc and _group:
                        rdxdr_d1_d2 = np.einsum('ik,ij->ijk', _rdxdr[:, d1, :], _x[:, d1+1:]) + \
                                      np.einsum('ijk,i->ijk', _rdxdr[:, d1+1:, :], _x[:, d1])
                    
                    # Append for x, dxdr, rdxdr
                    x[:, dcount] += x_square[:, d1]
                    if fc:
                        dxdr[:, dcount, :] += dxdr_square[:, d1, :]
                    if sc and _group:
                        rdxdr[:, dcount, :] += rdxdr_square[:, d1, :]

                    dcount += 1
                    
                    x[:, dcount:dcount+len(x_d1_d2[0])] += x_d1_d2
                    if fc:
                        dxdr[:, dcount:dcount+len(x_d1_d2[0]), :] += dxdr_d1_d2
                    if sc and _group:
                        rdxdr[:, dcount:dcount+len(x_d1_d2[0]), :] += rdxdr_d1_d2
                    dcount += len(x_d1_d2[0])
                
            else:
                x = data['x'][:, :d_max]
                if fc:
                    seq = data['seq']   
                    dxdr = data['dxdr'][:, :d_max, :]
                if sc and _group:
                    rdxdr = data['rdxdr'][:, :d_max, :]
            
            elements = data['elements']
            
            # Arranging x and dxdr for energy and forces.
            bias_weights = 1.0
            
            if train:
                sna = np.zeros([len(self.elements), 1+d_max])
                if fc:
                    snad = np.zeros([len(self.elements), len(x), 1+d_max, 3])
                if sc and _group:
                    snav = np.zeros([len(self.elements), 1+d_max, 6])
                
                _sna, _snad, _snav, _count = {}, {}, {}, {}
                for element in self.elements:
                    _sna[element] = None
                    _snad[element] = None
                    _snav[element] = None
                    _count[element] = 0
                
                # Loop over the number of atoms in a structure.
                for e, element in enumerate(elements):
                    if _sna[element] is None:
                        _sna[element] = 1 * x[e]
                        if fc:
                            shp = snad.shape
                            _snad[element] = np.zeros([shp[1], shp[2]-1, shp[3]])
                            arr = np.where(seq[:, 0]==e)[0]
                            _seq = seq[arr][:, 1]
                            _snad[element][_seq] = -1 * dxdr[arr]
                        if sc and _group:
                            _snav[element] = -1 * rdxdr[e]  # [d, 6]
                    else:
                        _sna[element] += x[e]
                        if fc:
                            arr = np.where(seq[:, 0]==e)[0]
                            _seq = seq[arr][:, 1]
                            _snad[element][_seq] -= dxdr[arr]
                        if sc and _group: 
                            _snav[element] -= rdxdr[e]
                    _count[element] += 1

                for e, element in enumerate(self.elements):
                    if _count[element] > 0:
                        #_sna[element] /= _count[element]
                        sna[e, :] += np.hstack(([bias_weights*_count[element]], _sna[element]))
                        if fc:
                            snad[e, :, 1:, :] += _snad[element]
                        if sc and _group:
                            snav[e, 1:, :] += _snav[element]
                        
                # X for energy
                X[xcount, :] += sna.ravel()
                xcount += 1

                # X for forces.
                if fc:
                    for j in range(snad.shape[1]):
                        for k in range(snad.shape[3]):
                            X[xcount, :] += snad[:, j, :, k].ravel()
                            xcount += 1
                
                # X for stress.
                if sc and _group: 
                    shp = snav.shape
                    X[xcount:xcount+6, :] = snav.reshape([shp[0]*shp[1], shp[2]]).T
                    xcount += 6

            else:
                if fc:
                    snad = np.zeros([len(self.elements), len(x), 1+d_max, 3])
                if sc and _group:
                    snav = np.zeros([len(self.elements), 1+d_max, 6])
                _snad, _snav, _count = {}, {}, {}
                for element in self.elements:
                    _snad[element] = None
                    _snav[element] = None
                    _count[element] = 0

                # Loop over the number of atoms in a structure.
                for e, element in enumerate(elements):
                    elem_cnt = self.elements.index(element)
                    X[xcount, elem_cnt*(1+d_max):(elem_cnt+1)*(1+d_max)] = np.hstack(([bias_weights], x[e]))
                    xcount += 1
                    if _snad[element] is None:
                        if fc:
                            shp = snad.shape
                            _snad[element] = np.zeros([shp[1], shp[2]-1, shp[3]])
                            arr = np.where(seq[:, 0]==e)[0]
                            _seq = seq[arr][:, 1]
                            _snad[element][_seq] = -1 * dxdr[arr]
                        if sc and _group:
                            _snav[element] = -1 * rdxdr[e]  # [d, 6]
                    else:
                        if fc:
                            arr = np.where(seq[:, 0]==e)[0]
                            _seq = seq[arr][:, 1]
                            _snad[element][_seq] -= dxdr[arr]
                        if sc and _group: 
                            _snav[element] -= rdxdr[e]
                    _count[element] += 1

                for e, element in enumerate(self.elements):
                    if _count[element] > 0:
                        if fc:
                            snad[e, :, 1:, :] += _snad[element]
                        if sc and _group:
                            snav[e, 1:, :] += _snav[element]

                # X for forces.
                if fc:
                    for j in range(snad.shape[1]):
                        for k in range(snad.shape[3]):
                            X[xcount, :] += snad[:, j, :, k].ravel()
                            xcount += 1
                
                # X for stress.
                if sc and _group: 
                    shp = snav.shape
                    X[xcount:xcount+6, :] = snav.reshape([shp[0]*shp[1], shp[2]]).T
                    xcount += 6
        
        if train:
            db.close()

        return X


    def parse_features(self, data):
        """ Parse features (energy, forces, and stress) into 1-D array.
        
        Returns
        -------
        y: 1-D array [n+m*3+n*3*3,]
            y contains the energy, forces, and stress of structures 
            in 1-D array. Force and stress may not be present.
            
            n = # of structures
            m = # of atoms in a unit cell

        w: 1-D array [n+m*3+n*3*3,]
            w contains the relative importance between energy, forces, 
            and stress.
        """
        db = shelve.open(self.path+data)
        self.no_of_atoms = 0
        self.stress_components = 0

        y = None # store the features (energy+forces+stress)
        w = None # weight of each sample
        
        for i in range(self.no_of_structures):
            data = db[str(i)]
            #energy = np.array([data['energy']/len(data['elements'])])
            energy = np.array([data['energy']])
            w_energy = np.array([1.])

            if self.force_coefficient:
                force = np.array(data['force']).ravel()
                w_force = np.array([self.force_coefficient]*len(force))

                if self.stress_coefficient and (data['group'] in self.stress_group):     # energy + forces + stress
                    stress = np.array(data['stress'])#.flat[[0,3,5,3,1,4,5,4,2]]
                    w_stress = np.array([self.stress_coefficient]*len(stress))
                    self.stress_components += 6
                    
                    if y is None:
                        y = np.concatenate((energy, force, stress))
                        w = np.concatenate((w_energy, w_force, w_stress))
                    else:
                        y = np.concatenate((y, energy, force, stress))
                        w = np.concatenate((w, w_energy, w_force, w_stress))
                else:                                                                           # energy + forces
                    if y is None:
                        y = np.concatenate((energy, force))
                        w = np.concatenate((w_energy, w_force))
                    else:
                        y = np.concatenate((y, energy, force))
                        w = np.concatenate((w, w_energy, w_force))
                
                # Count the number of atoms for the entire structures.
                self.no_of_atoms += len(data['force'])

            else:
                if self.stress_coefficient and (data['group'] in self.stress_group):    # energy + stress
                    stress = np.array(data['stress'])#.flat[[0,3,5,3,1,4,5,4,2]]
                    w_stress = np.array([self.stress_coefficient]*len(stress))
                    self.stress_components += 6
                    
                    if y is None:
                        y = np.concatenate((energy, stress))
                        w = np.concatenate((w_energy, w_stress))
                    else:
                        y = np.concatenate((y, energy, stress))
                        w = np.concatenate((w, w_energy, w_stress))

                else:                                                                           # energy only
                    if y is None:
                        y = energy
                        w = w_energy
                    else:
                        y = np.concatenate((y, energy))
                        w = np.concatenate((w, w_energy))
        
        db.close()
        gc.collect()

        return y, w



import numpy as np
from ase import Atoms
from phonopy.api_phonopy import Phonopy

class Phonon(object):
    """
    This module is to compute phonon from the trained Pyxtal_FF model.
    We use phonopy API to construct and handle force constant matrix

    Args:
    atoms: ase atoms object
    calculator: ase calculator object
    supercell_matrix:  3x3 matrix supercell
    primitive cell:  3x3 matrix primitive cell
    displacement_distance: displacement distance in Angstroms
    show_progress: Set true to display progress of calculation
    """

    def __init__(self,
                 atoms,
                 calculator,
                 supercell_matrix=np.identity(3),
                 primitive_matrix='auto',
                 displacement_distance=0.01,
                 show_progress=True,
                 symmetrize=True):

        self._structure = atoms
        self._calculator = calculator
        self._supercell_matrix = supercell_matrix
        self._primitive_matrix = primitive_matrix
        self._d = displacement_distance
        self._show_progress = show_progress
        self._symmetrize = symmetrize

        self._force_constants = None
        self._data_set = None

    def get_force_constants(self, include_data_set=False):
        """
        calculate the force constants with phonopy 

        Return: 
        ForceConstants type object containing force constants
        """

        self.phonon = Phonopy(self._structure,
                              self._supercell_matrix,
                              primitive_matrix = self._primitive_matrix,
                              is_symmetry = self._symmetrize)
        self.phonon.get_displacement_dataset()
        self.phonon.generate_displacements(distance=self._d) 
        cells_with_disp = self.phonon.get_supercells_with_displacements()
        data_set = self.phonon.get_displacement_dataset()
        # Compute forces for the displaced structures
        for i, cell in enumerate(cells_with_disp):
            if self._show_progress:
                print('displacement {} / {}'.format(i+1, len(cells_with_disp)))
            forces = self.get_forces(cell)
            data_set['first_atoms'][i]['forces'] = forces

        self.phonon.set_displacement_dataset(data_set)
        self.phonon.produce_force_constants()
        self._force_constants = self.phonon.get_force_constants()
        self._data_set = data_set

        return self._force_constants

    def get_forces(self, cell_with_disp):
        """
        Calculate the forces of a supercell using lammps
        Args:
        cell_with_disp: supercell from which determine the forces

        Return:
        forces: [Natoms_of_supercell, 3] array
        """

        numbers = self._structure.numbers
        s_numbers = []
        mult = np.linalg.det(self._supercell_matrix)
        atom_id = 0
        disps = cell_with_disp.get_positions()
        for id, disp in enumerate(disps):
            s_numbers.append(numbers[atom_id])
            if id%mult == (mult-1):
                atom_id += 1
        
        cell = self._structure.cell.dot(self._supercell_matrix)
        supercell = Atoms(s_numbers, positions=disps, cell=cell, pbc=[1,1,1])
        supercell.set_calculator(self._calculator)
        return supercell.get_forces()

if  __name__ == "__main__":

    from optparse import OptionParser
    from ase.build import bulk
    from pyxtal_ff.calculator import PyXtalFFCalculator
    from phonopy.phonon.band_structure import get_band_qpoints_and_path_connections

    parser = OptionParser()
    parser.add_option("-f", "--file", dest="file",
                      help="pretrained file from pyxtal_ff, REQUIRED",
                      metavar="file")

    (options, args) = parser.parse_args()
    calc = PyXtalFFCalculator(mliap=options.file, logo=False)
    si = bulk('Si', 'diamond', a=5.459, cubic=True)
    supercell_matrix=np.diag([2,2,2])
    ph = Phonon(si, calc, supercell_matrix=supercell_matrix) 
    force_constants = ph.get_force_constants()
    ph.phonon.set_force_constants(force_constants)

    #ph.phonon.auto_band_structure()
    path = [[[0, 0, 0], [0.5, 0, 0.5], [0.375, 0.375, 0.75], [0, 0, 0], [0.5, 0.5, 0.5]]]
    labels = ["$\\Gamma$", "X", "K", "$\\Gamma$", "L"]
    qpoints, connections = get_band_qpoints_and_path_connections(path, npoints=51)
    ph.phonon.run_band_structure(qpoints, path_connections=connections, labels=labels)

    ph.phonon.set_mesh([10,10,10])
    ph.phonon.set_total_DOS()
    ph.phonon.plot_band_structure_and_dos().savefig('Si_phonon.png', dpi=300)
    print('The detected space group is {:s} with a tolerance of {:.2e}'.format(
          ph.phonon._symmetry.dataset['international'], ph.phonon._symprec))



import os
import torch
from pyxtal_ff.version import __version__
from pyxtal_ff.utilities import Database
from pyxtal_ff.models.polynomialregression import PR
from pyxtal_ff.models.neuralnetwork import NeuralNetwork

class PyXtal_FF():
    def __init__(self, descriptors=None, model=None, logo=True):
        """ PyXtal_FF develops Machine Learning Interatomic Potential.
        
        Parameters
        ----------
        descriptors: dict
            The atom-centered descriptors parameters are defined here.
            
            The list of the descriptors keys:
            - type: str (SO4)
                The type of atom-centered descriptors.
                + ACSF (BehlerParrinello Gaussian symmetry)
                + wACSF (weighted Gaussian symmetry)
                + EAD (embeded atom density)
                + SO4 (bispectrum)
                + SO3 (smoothed powerspectrum)
                + SNAP (similar to SO4 but the weighting and Rc schemes are adopted from LAMMPS)
            - Rc: float/dictionary
                The radial cutoff of the descriptors. Dictionary form is particularly for SNAP.
            - weights: dictionary
                The relative species weights.
            - N_train: int
                The number of crystal structures in training data set 
                to be converted into descriptors.
            - N_test: int
                The number of crystal structures in test data set 
                to be converted into descriptors.
            - ncpu: int
                The number of cpu core to use for converting crystal structures 
                into descriptors.
            - stress: bool (False)
                Compute rdxdr (needed for stress calculation) or not
            - force: bool (True)
                Compute dxdr (needed for force calculation) or not
            - cutoff: str
                The cutoff function.
            - parameters: dict
                Example,
                + BehlerParrinello
                    {'G2': {'eta': [.3, 2.], 'Rs': [1., 2.]}, 
                     'G4': {'eta': [.3, .7], 'lambda': [-1, 1], 'zeta': [.8, 2]}}
                + EAD
                    {'L': 2, 'eta': [.3, .7], 'Rs': [.1, .2]}
                + SO4/Bispectrum
                    {'lmax': 3}
                + SO3
                    {'nmax': 1, 'lmax': 3, 'alpha': 2.0}
                + SNAP
                    {'weights': {'Si': 1.0, 'O': 2.0},
                     'Rc': {'Si': 4.0, 'O': 5.0}
            - zbl: dict
                {'inner': 4.0, 'outer': 4.5}

        model: dict
            Machine learning parameters are defined here.

            The list of the model keys:
            - algorithm: str (*NN and *PR)
                The desired machine learning algorithm for potential 
                development. Choose between ['PolynomialRegression', 'PR'] or
                ['NeuralNetwork', 'NN'].
            - system: list of str (*NN, *PR, and *GPR)
                A list of atomic species in the crystal system.
                e.g. ['Si', 'O']
            - hiddenlayers: list of int (*NN)
                [3, 3] contains 2 hidden layers with 3 nodes each.
            - activation: str or list of str (*NN)
                The activation function for the neural network model.
                Currently, there are tanh, sigmoid, and linear.
            - random_seed: int (*NN)
                If the Neural Network is initialized from the beginning,
                the random_seed is used to established the initial 
                Neural Network weights randomly.
            - batch_size: int (*NN)
                batch_size is used for online learning. The weights of the 
                Neural Network is updated for each batch_size.
            - epoch: int (*NN and *GPR)
                A measure of the number of times all of the training vectors 
                are used once to update the weights.
            - device: str (*NN and *GPR)
                The device used to train: 'cpu' or 'cuda'.
            - force_coefficient: float (*NN, *PR, and *GPR)
                This parameter is used as the penalty parameter to scale 
                the force contribution relative to the energy.
            - stress_coefficient: float (*NN, *PR, and *GPR)
                This parameter is used as the balance parameter scaling
                the stress contribution relative to the energy.
            - stress_group: list of strings, not bool! (*NN, *PR, and *GPR)
                Only the intended group will be considered in stress training.
            - softmax_beta: float (*NN)
                The parameters for Softmax Energy Penalty function.
            - unit: str (*NN)
                The unit of energy ('eV' or 'Ha'). 
                The default unit of energy is 'eV'. If 'Ha' is used,
                Bohr is the unit length; otherwise, Angstrom is used.
            - restart: str (*NN)
                To continue Neural Network training from where it was left off.
            - optimizer: dict (*NN and *GPR)
                Define the optimization method used to update NN parameters.
            - path: str (*NN, *PR, *GPR)
                The user defined path to store the NN results.
                path has to be ended with '/'.
            - memory: str (*NN)
                There are two options: 'in' or 'out'. 'in' will use load all
                descriptors to memory as 'out' will call from disk as needed.
            - order: int (*PR)
                Order is used to determined the polynomial order. 
                For order = 1, linear is employed, and quadratic is employed 
                for order = 2.
            - d_max: int (*PR)
                The maximum number of descriptors used.
            - alpha: float (*NN and *PR)
                L2 penalty (regularization term) parameter.
            - norm: int (*PR)
                This argument defines a model to calculate the regularization
                term. It takes only 1 or 2 as its value: Manhattan or Euclidean
                norm, respectively. If alpha is None, norm is ignored.
            - noise: float (*GPR)
                The noise added to the Gaussian Kernel
            - kernel: str (*GPR)
                The kernel specifying the covariance function of the GPR.
                The current development allows "RBF" and "DotProduct".
                
        (*) required.
        (*NN) for Neural Network algorithm only.
        (*PR) for Polynomial Regression algorithm only.
        (*GPR) for Gaussian Process Regressor algorithm only.
        """
        if logo:
            self.print_logo()
        
        # Checking the keys in descriptors
        descriptors_keywords = ['type', 'Rc', 'weights', 'N_train', 'N_test', 'cutoff',
                                'force', 'stress', 'ncpu', 'parameters', 'base_potential']
        if descriptors is not None:
            for key in descriptors.keys():
                if key not in descriptors_keywords:
                    msg = f"Don't recognize {key} in descriptors. "+\
                          f"Here are the keywords: {descriptors_keywords}."
                    raise NotImplementedError(msg)

        # Set up default descriptors parameters
        self._descriptors = {'system': model['system'],
                             'type': 'Bispectrum',
                             'Rc': 5.0,
                             'weights': None,
                             'N': None,
                             'N_train': None,
                             'N_test': None,
                             'ncpu': 1,
                             'force': True,
                             'stress': True,
                             'cutoff': 'cosine',
                             'base_potential': False,
                             }
        
        # Update the default based on user-defined descriptors
        if descriptors is not None:
            self._descriptors.update(descriptors)
            if 'type' in descriptors and descriptors['type'] in ['EAD', 'ead']:
                _parameters = {'L': 3, 'eta': [0.1], 'Rs': [1.]}
            elif 'type' in descriptors and descriptors['type'] in ['SO3', 'SOAP']:
                _parameters = {'nmax': 3, 'lmax': 3, 'alpha': 2.0}
            else:
                _parameters = {'lmax': 3, 'rfac': 0.99363, 'normalize_U': False}
            if 'parameters' in descriptors:
                _parameters.update(descriptors['parameters'])
                self._descriptors['parameters'] = _parameters

            # Check for the SNAP type
            if self._descriptors['type'] in ['SNAP', 'snap']:
                if not isinstance(self._descriptors['weights'], dict):
                    msg = "The weights for SNAP type must be defined as a dictionary."
                    raise ValueError(msg)
                #if not isinstance(self._descriptors['Rc'], dict):
                #    msg = "The Rc for SNAP type must be defined as a dictionary."
                #    raise ValueError(msg)

        # Create new directory to dump all the results.
        # E.g. for default 'Si-O-Bispectrum/'
        if 'path' in model:
            self.path = model['path']
        else:
            _system = model['system']
            self.path = "-".join(_system) + "-"
            self.path += self._descriptors['type'] + "/"

        if logo:
            if not os.path.exists(self.path):
                os.mkdir(self.path)
            self.print_descriptors(self._descriptors)
        
        # Checking the keys in model.
        keywords = ['algorithm', 'system', 'hiddenlayers', 'activation', 
                    'random_seed', 'force_coefficient', 'unit', 'softmax_beta', 
                    'restart', 'optimizer', 'path', 'order', 'd_max', 
                    'epoch', 'device', 'alpha', 'batch_size', 'noise', 'kernel',
                    'norm', 'stress_coefficient', 'stress_group', 'memory']
        for key in model.keys():
            if key not in keywords:
                msg = f"Don't recognize {key} in model. "+\
                      f"Here are the keywords: {keywords}."
                raise NotImplementedError(msg)
        
        # Create model
        pr_keywords = ['PolynomialRegression', 'PR']
        nn_keywords = ['NeuralNetwork', 'NN']
        if 'algorithm' not in model:
            model['algorithm'] = 'NN'

        if model['algorithm'] in pr_keywords:
            self.algorithm = 'PR'
        elif model['algorithm'] in nn_keywords:
            self.algorithm = 'NN'
        else:
            msg = f"{model['algorithm']} is not implemented."
            raise NotImplementedError(msg)
        
        self._model = model

    def todict(self):
        return {"descriptor": self._descriptors, "model": self._model}

    def run(self, mode='train', TrainData=None, TestData=None, mliap=None):
        """ Command PyXtal_FF to run in 2 modes:
        1. train
            In train mode, PyXtal_FF needs TrainData and/or TestData to be defined.
            
            TrainData: str
                TrainData indicates the location the training data set file.
                After the training of Neural Network is over, TrainData will be 
                self-evaluated for the accuracy.
            
            TestData: str
                TestData is an optional argument. If the test data set is present, 
                PyXtal_FF will evaluate the accuracy of the developed potential 
                with the TestData.
        

        2. predict
            In predict mode, PyXtal_FF need the saved machine learning interatomic
            potential.

            mliap: str
                The machine learning interatomic potential.
        """
        if mode == 'train':
            assert TrainData is not None, "TrainData can't be None for train mode."

            # Instantiate model
            self._MODEL(self._model)

            # Calculate descriptors.
            self._descriptors.update({'N': self._descriptors['N_train']})
            if not os.path.exists(self.path+'Train_db.dat') and not os.path.exists(self.path+'Train_db.db'):
                trainDB = Database(name=self.path+'Train_db')
                trainDB.store(TrainData, self._descriptors, True, self.path+'ase.db')
            else:
                trainDB = Database(name=self.path+'Train_db')
                trainDB.store(TrainData, self._descriptors, False)
            trainDB.close()

            if TestData is not None:
                EvaluateTest = True
                self._descriptors.update({'N': self._descriptors['N_test']}) 
                if not os.path.exists(self.path+'Test_db.dat'):
                    testDB = Database(name=self.path+'Test_db')
                    testDB.store(TestData, self._descriptors, True, self.path+'ase.db')
                else:
                    testDB = Database(name=self.path+'Test_db')
                    testDB.store(TestData, self._descriptors, False)
                testDB.close()

            else:
                EvaluateTest = False
            
            print("=========================== Training =============================\n")

            self.model.train('Train_db', optimizer=self.optimizer)
            self.model.save_checkpoint(des_info=self._descriptors)
            
            print("==================================================================\n")
            
            print(f"==================== Evaluating Training Set ====================\n")

            train_stat = self.model.evaluate('Train_db', figname='Train.png')
            
            print("==================================================================\n")

            if EvaluateTest:
                print("================= Evaluating Testing Set =====================\n")

                test_stat =  self.model.evaluate('Test_db', figname='Test.png')
                
                print("==============================================================\n")
            else:
                test_stat = None

            return (train_stat, test_stat)
        
        elif mode == 'predict':
            self._model['algorithm'] = torch.load(mliap)['algorithm']
            self.algorithm = self._model['algorithm']
            self._MODEL(self._model)
            self._descriptors = self.model.load_checkpoint(filename=mliap)

    
    def _MODEL(self, model):
        """ Machine learning model is created here. """
                    
        if self.algorithm == 'NN':
            _model = {'system': None,
                      'hiddenlayers': [6, 6],
                      'activation': ['Tanh', 'Tanh', 'Linear'],
                      'random_seed': None,
                      'epoch': 100,
                      'batch_size': None,
                      'device': 'cpu',
                      'force_coefficient': 0.03,
                      'stress_coefficient': None,
                      'stress_group': None,
                      'softmax_beta': None,
                      'alpha': None,
                      'unit': 'eV',
                      'restart': None,
                      'path': self.path,
                      'memory': 'in',
                      'optimizer': {},
                      }
            _model.update(model)
            
            if len(_model['activation']) != len(_model['hiddenlayers']) + 1:
                msg = '\nWarning: Incompatible activation functions and hiddenlayers.'
                print(msg)
                print('hiddenlayers: ', _model['hiddenlayers'])
                print('activations: ', _model['activation'])
                _model['activation'] = ['Tanh']*len(model['hiddenlayers'])+['Linear']
                print('revised activations: ', _model['activation'])
    
            if 'parameters' not in _model['optimizer']:
                _model['optimizer']['parameters'] = {}
            if 'derivative' not in _model['optimizer']:
                _model['optimizer']['derivative'] = True
            if 'method' not in _model['optimizer']:
                _model['optimizer']['method'] = 'lbfgs'

                # In full batch-LBFGS, epoch is 1. 
            if _model['optimizer']['method'] in ['lbfgs', 'LBFGS', 'lbfgsb']:
                if _model['batch_size'] is None: #full batch
                    _model['optimizer']['parameters']['max_iter'] = _model['epoch']
                    _model['epoch'] = 1
                else:
                    _model['optimizer']['parameters']['max_iter'] = 20

            self.model = NeuralNetwork(elements=_model['system'],
                                       hiddenlayers=_model['hiddenlayers'],
                                       activation=_model['activation'],
                                       random_seed=_model['random_seed'],
                                       epoch=_model['epoch'],
                                       batch_size=_model['batch_size'],
                                       device=_model['device'],
                                       alpha=_model['alpha'],
                                       force_coefficient=_model['force_coefficient'],
                                       stress_coefficient=_model['stress_coefficient'],
                                       stress_group=_model['stress_group'],
                                       softmax_beta=_model['softmax_beta'],
                                       unit=_model['unit'],
                                       restart=_model['restart'],
                                       path=_model['path'],
                                       memory=_model['memory'])
            self.optimizer = _model['optimizer']
                
        elif self.algorithm == 'PR':
            _model = {'system': None,
                      'force_coefficient': 0.0001,
                      'stress_coefficient': None,
                      'stress_group': None,
                      'order': 1,
                      'path': self.path,
                      'alpha': None,
                      'norm': 2,
                      'd_max': None,
                      }
            _model.update(model)
            self.model = PR(elements=_model['system'],
                            force_coefficient=_model['force_coefficient'],
                            stress_coefficient=_model['stress_coefficient'],
                            stress_group=_model['stress_group'],
                            order=_model['order'],
                            path=_model['path'],
                            alpha=_model['alpha'],
                            norm=_model['norm'],
                            d_max=_model['d_max'])
            self.optimizer = None


    def print_descriptors(self, _descriptors):
        """ Print the descriptors information. """

        print('Descriptor parameters:')
        keys = ['type', 'Rc', 'cutoff']
        for key in keys:
            print('{:12s}: {:}'.format(key, _descriptors[key]))

        if _descriptors['type'] in ['SO4', 'Bispectrum']:
            key_params = ['lmax', 'normalize_U']
        elif _descriptors['type'] in ['SO3', 'SOAP']:
            key_params = ['nmax', 'lmax', 'alpha']
        elif _descriptors['type'] in ['SNAP', 'snap']:
            key_params = ['lmax', 'rfac']
        elif _descriptors['type'] == 'EAD':
            key_params = ['L', 'eta', 'Rs']
        else:
            key_params = []

        for key in key_params:
            print('{:12s}: {:}'.format(key, _descriptors['parameters'][key]))
        print('\n')


    def print_logo(self):
        """ Print PyXtal_FF logo and version. """
        
        print("\n")
        print("""
               ______       _    _          _         _______ _______ 
              (_____ \     \ \  / /        | |       (_______|_______)
               _____) )   _ \ \/ / |_  ____| |        _____   _____   
              |  ____/ | | | )  (|  _)/ _  | |       |  ___) |  ___)  
              | |    | |_| |/ /\ \ |_( ( | | |_______| |     | |      
              |_|     \__  /_/  \_\___)_||_|_(_______)_|     |_|      
                     (____/      """)
        print("\n")
        print('        A Python package for Machine Learning Interatomic Force Field')
        print('         Developed by Zhu\'s group at University of Nevada Las Vegas')
        print('    The source code is available at https://github.com/qzhu2017/FF-project')
        print("\n")
        print('=========================== version', __version__,'=============================\n')



#python -m unittest pyxtal_ff/test_all.py
import shutil
import unittest
import numpy as np
from ase import Atoms
from ase.build import bulk, sort
from ase.cluster.cubic import FaceCenteredCubic
from ase import units
from ase.optimize import BFGS

from pkg_resources import resource_filename
from pyxtal_ff import PyXtal_FF
from pyxtal_ff.calculator import PyXtalFFCalculator, optimize
from pyxtal_ff.calculator.elasticity import fit_elastic_constants
from pyxtal_ff.descriptors.SO3 import SO3
from pyxtal_ff.descriptors.EAD import EAD
from pyxtal_ff.descriptors.ACSF import ACSF
from pyxtal_ff.descriptors.SO4 import SO4_Bispectrum as SO42
from pyxtal_ff.descriptors.SNAP import SO4_Bispectrum as SO41
from pyxtal_ff.utilities.base_potential import ZBL
np.set_printoptions(formatter={'float': '{: 12.4f}'.format})

def get_rotated_struc(struc, angle=0, axis='x'):
    s_new = struc.copy()
    s_new.rotate(angle, axis)
    cell = 17.22*np.eye(3)
    p_struc = Atoms(s_new.symbols.numbers, positions=s_new.positions, cell=cell, pbc=True)
    return p_struc

def get_perturbed_struc(struc, p0, p1, eps):
    s_new = struc.copy()
    pos = s_new.positions
    pos[p0, p1] += eps
    cell = 17.22*np.eye(3)
    p_struc = Atoms(s_new.symbols.numbers, positions=pos, cell=cell, pbc=True)
    return p_struc

# NaCl Cluster
nacl = bulk('NaCl', crystalstructure='rocksalt', a=5.691694, cubic=True)
nacl = sort(nacl, tags=[0,4,1,5,2,6,3,7])
nacl.set_pbc((0,0,0))
nacl = get_rotated_struc(nacl, angle=1)

# Descriptors Parameters
eps = 1e-8
rc = 6.00
nmax, lmax = 2, 2
ead_params = {'L': lmax,
              'eta': [0.36, 0.036],
              'Rs': [1.2, 2.1]}
acsf_params = {'G2': {'eta': [0.36, 0.036],
                      'Rs': [1.2, 2.1]},
               'G4': {'eta': [0.36, 0.036],
                      'Rs': [1.2, 2.1],
                      'lambda': [-1, 1],
                      'zeta': [1.0, 1.5]},
               'G5': {'eta': [0.18, 0.018],
                      'Rs': [1.1, 2.3],
                      'lambda': [-1, 1],
                      'zeta': [1.0, 1.7]}}

# Neural Network Parameters
TrainData = resource_filename("pyxtal_ff", "datasets/Si/PyXtal/Si4.json")
parameters = {'lmax': 2}
system = ['Si']
descriptor = {'type': 'Bispectrum',
              'parameters': parameters,
              'Rc': 3.0,
              'N_train': 10,
              'stress': True,
              }

descriptor_comp = {'type': 'Bispectrum',
                   'parameters': parameters,
                   'Rc': 3.0,
                   'N_train': 10,
                  }

# model file
bp_model = resource_filename("pyxtal_ff", "datasets/Si/PyXtal/bp-16-16-checkpoint.pth")

class TestEAD(unittest.TestCase):
    struc = get_rotated_struc(nacl)
    rho0 = EAD(parameters=ead_params, Rc=rc, derivative=True).calculate(struc)
    struc = get_rotated_struc(nacl, 10, 'x')
    rho1 = EAD(parameters=ead_params, Rc=rc, derivative=True).calculate(struc)

    def test_rho_rotation_variance(self):
        array1 = self.rho0['x']
        array2 = self.rho1['x']
        self.assertTrue(np.allclose(array1, array2))

    def test_drhodR_rotation_variance(self):
        for i in range(len(self.rho0['x'])):
            array1 = np.linalg.norm(self.rho0['dxdr'][i,:,:], axis=1)
            array2 = np.linalg.norm(self.rho1['dxdr'][i,:,:], axis=1)
            self.assertTrue(np.allclose(array1, array2))

    def test_drhodR_vs_numerical(self):
        shp = self.rho0['x'].shape
        array1 = np.zeros([shp[0], shp[0], shp[1], 3])
        for _m in range(shp[0]):
            ids = np.where(self.rho0['seq'][:,1]==_m)[0]
            array1[self.rho0['seq'][ids, 0], _m, :, :] += self.rho0['dxdr'][ids,:,:]
        for j in range(shp[0]):
            for k in range(3):
                struc = get_perturbed_struc(nacl, j, k, eps)
                rho2 = EAD(parameters=ead_params, Rc=rc, derivative=False).calculate(struc)
                array2 = (rho2['x'] - self.rho0['x'])/eps
                self.assertTrue(np.allclose(array1[:,j,:,k], array2, atol=1e-4))

class TestACSF(unittest.TestCase):
    struc = get_rotated_struc(nacl)
    ACSF0 = ACSF(symmetry_parameters=acsf_params, Rc=rc, derivative=True).calculate(struc)
    struc = get_rotated_struc(nacl, 10, 'x')
    ACSF1 = ACSF(symmetry_parameters=acsf_params, Rc=rc, derivative=True).calculate(struc)

    def test_ACSF_rotation_variance(self):
        array1 = self.ACSF0['x']
        array2 = self.ACSF1['x']
        self.assertTrue(np.allclose(array1, array2))

    def test_dACSFdR_rotation_variance(self):
        for i in range(len(self.ACSF0['x'])):
            array1 = np.linalg.norm(self.ACSF0['dxdr'][i,:,:], axis=1)
            array2 = np.linalg.norm(self.ACSF1['dxdr'][i,:,:], axis=1)
            self.assertTrue(np.allclose(array1, array2))

    def test_dACSFdR_vs_numerical(self):
        shp = self.ACSF0['x'].shape
        array1 = np.zeros([shp[0], shp[0], shp[1], 3])
        for _m in range(shp[0]):
            ids = np.where(self.ACSF0['seq'][:,1]==_m)[0]
            array1[self.ACSF0['seq'][ids, 0], _m, :, :] += self.ACSF0['dxdr'][ids,:,:]
        
        for j in range(shp[0]):
            for k in range(3):
                struc = get_perturbed_struc(nacl, j, k, eps)
                ACSF2 = ACSF(symmetry_parameters=acsf_params, Rc=rc, derivative=False).calculate(struc)
                array2 = (ACSF2['x'] - self.ACSF0['x'])/eps
                self.assertTrue(np.allclose(array1[:,j,:,k], array2, atol=1e-6))

class TestwACSF(unittest.TestCase):
    struc = get_rotated_struc(nacl)
    wACSF0 = ACSF(symmetry_parameters=acsf_params, Rc=rc, derivative=True, atom_weighted=True).calculate(struc)
    struc = get_rotated_struc(nacl, 10, 'x')
    wACSF1 = ACSF(symmetry_parameters=acsf_params, Rc=rc, derivative=True, atom_weighted=True).calculate(struc)

    def test_wACSF_rotation_variance(self):
        array1 = self.wACSF0['x']
        array2 = self.wACSF1['x']
        self.assertTrue(np.allclose(array1, array2))

    def test_dwACSFdR_rotation_variance(self):
        for i in range(len(self.wACSF0['x'])):
            array1 = np.linalg.norm(self.wACSF0['dxdr'][i,:,:], axis=1)
            array2 = np.linalg.norm(self.wACSF1['dxdr'][i,:,:], axis=1)
            self.assertTrue(np.allclose(array1, array2))

    def test_dwACSFdR_vs_numerical(self):
        shp = self.wACSF0['x'].shape
        array1 = np.zeros([shp[0], shp[0], shp[1], 3])
        for _m in range(shp[0]):
            ids = np.where(self.wACSF0['seq'][:,1]==_m)[0]
            array1[self.wACSF0['seq'][ids, 0], _m, :, :] += self.wACSF0['dxdr'][ids,:,:]
        
        for j in range(shp[0]):
            for k in range(3):
                struc = get_perturbed_struc(nacl, j, k, eps)
                wACSF2 = ACSF(symmetry_parameters=acsf_params, Rc=rc, derivative=False, atom_weighted=True).calculate(struc)
                array2 = (wACSF2['x'] - self.wACSF0['x'])/eps
                self.assertTrue(np.allclose(array1[:,j,:,k], array2, atol=1e-6))

class TestSO3(unittest.TestCase):
    struc = get_rotated_struc(nacl)
    p0 = SO3(nmax=nmax, lmax=lmax, rcut=rc, derivative=True).calculate(struc)
    struc = get_rotated_struc(nacl, 10, 'x')
    p1 = SO3(nmax=nmax, lmax=lmax, rcut=rc, derivative=True).calculate(struc)

    def test_SO3_rotation_variance(self):
        array1 = self.p0['x']
        array2 = self.p1['x']
        self.assertTrue(np.allclose(array1, array2))

    def test_dPdR_rotation_variance(self):
        for i in range(len(self.p0['x'])):
            array1 = np.linalg.norm(self.p0['dxdr'][i,:,:], axis=1)
            array2 = np.linalg.norm(self.p1['dxdr'][i,:,:], axis=1)
            self.assertTrue(np.allclose(array1, array2))

    def test_dPdR_vs_numerical(self):
        shp = self.p0['x'].shape
        array1 = np.zeros([shp[0], shp[0], shp[1], 3])
        for _m in range(shp[0]):
            ids = np.where(self.p0['seq'][:,1]==_m)[0]
            array1[self.p0['seq'][ids, 0], _m, :, :] += self.p0['dxdr'][ids,:,:]
        
        for j in range(shp[0]):
            for k in range(3):
                struc = get_perturbed_struc(nacl, j, k, eps)
                p2 = SO3(nmax=nmax, lmax=lmax, rcut=rc, derivative=False).calculate(struc)
                array2 = (p2['x'] - self.p0['x'])/eps
                self.assertTrue(np.allclose(array1[:,j,:,k], array2, atol=1e-4))

class TestSNAP(unittest.TestCase):
    struc = get_rotated_struc(nacl)
    SNAP0 = SO41(weights={'Na':0.3,'Cl':0.7}, lmax=lmax, rcut=rc, rfac0=0.99363).calculate(struc)
    struc = get_rotated_struc(nacl, 10, 'x')
    SNAP1 = SO41(weights={'Na':0.3,'Cl':0.7}, lmax=lmax, rcut=rc, rfac0=0.99363).calculate(struc)

    def test_SNAP_rotation_variance(self):
        array1 = self.SNAP0['x']
        array2 = self.SNAP1['x']
        self.assertTrue(np.allclose(array1, array2))

    def test_dSNAPdR_rotation_variance(self):
        for i in range(len(self.SNAP0['x'])):
            array1 = np.linalg.norm(self.SNAP0['dxdr'][i,:,:], axis=1)
            array2 = np.linalg.norm(self.SNAP1['dxdr'][i,:,:], axis=1)
            self.assertTrue(np.allclose(array1, array2))

    def test_dSNAPdR_vs_numerical(self):
        shp = self.SNAP0['x'].shape
        array1 = np.zeros([shp[0], shp[0], shp[1], 3])
        for _m in range(shp[0]):
            ids = np.where(self.SNAP0['seq'][:,1]==_m)[0]
            array1[self.SNAP0['seq'][ids, 0], _m, :, :] += self.SNAP0['dxdr'][ids,:,:]
        
        for j in range(shp[0]):
            for k in range(3):
                struc = get_perturbed_struc(nacl, j, k, eps)
                SNAP2 = SO41(weights={'Na':0.3,'Cl':0.7}, lmax=lmax, rcut=rc, derivative=True, rfac0=0.99363).calculate(struc)
                array2 = (SNAP2['x'] - self.SNAP0['x'])/eps
                self.assertTrue(np.allclose(array1[:,j,:,k], array2, atol=1e-6))

class TestSO4(unittest.TestCase):
    struc = get_rotated_struc(nacl)
    so40 = SO42(lmax=lmax, rcut=rc, derivative=True).calculate(struc)
    struc = get_rotated_struc(nacl, 10, 'x')
    so41 = SO42(lmax=lmax, rcut=rc, derivative=True).calculate(struc)

    def test_so4_rotation_variance(self):
        array1 = self.so40['x']
        array2 = self.so41['x']
        self.assertTrue(np.allclose(array1, array2))

    def test_dso4dR_rotation_variance(self):
        for i in range(len(self.so40['x'])):
            array1 = np.linalg.norm(self.so40['dxdr'][i,:,:], axis=1)
            array2 = np.linalg.norm(self.so41['dxdr'][i,:,:], axis=1)
            self.assertTrue(np.allclose(array1, array2))

    def test_dso4dR_vs_numerical(self):
        shp = self.so40['x'].shape
        array1 = np.zeros([shp[0], shp[0], shp[1], 3])
        for _m in range(shp[0]):
            ids = np.where(self.so40['seq'][:,1]==_m)[0]
            array1[self.so40['seq'][ids, 0], _m, :, :] += self.so40['dxdr'][ids,:,:]
        
        for j in range(shp[0]):
            for k in range(3):
                struc = get_perturbed_struc(nacl, j, k, eps)
                so42 = SO42(lmax=lmax, rcut=rc, derivative=False).calculate(struc)
                array2 = (so42['x'] - self.so40['x'])/eps
                self.assertTrue(np.allclose(array1[:,j,:,k], array2, atol=1e-2))

class TestCalculator(unittest.TestCase):

    def testOptim(self):
        ff = PyXtal_FF(model={'system': ["Si"]}, logo=False)
        ff.run(mode='predict', mliap=bp_model)
        calc = PyXtalFFCalculator(ff=ff)
        si = bulk('Si', 'diamond', a=5.0, cubic=True)
        si.set_calculator(calc)
        si = optimize(si, box=True)
        self.assertTrue(abs(si.get_cell()[0][0]-5.469) <1e-2)

    def test_elastic(self):
        ff = PyXtal_FF(model={'system': ["Si"]}, logo=False)
        ff.run(mode='predict', mliap=bp_model)
        calc = PyXtalFFCalculator(ff=ff)
        si = bulk('Si', 'diamond', a=5.469, cubic=True)
        si.set_calculator(calc)
        C, C_err = fit_elastic_constants(si, symmetry='cubic', optimizer=BFGS)
        C /= units.GPa
        self.assertTrue(abs(C[0,0]-124.5)<1.0)

class TestZBL(unittest.TestCase):
    struc = nacl.copy()
    d1 = ZBL(2.0, 7.0).calculate(struc)
    d2 = ZBL(4.0, 4.5).calculate(struc)

    def test_ZBL(self):
        tenergy1 = 4.4444502 / len(self.struc)
        tforces1 = np.array(
                    [[-0.787367, -0.773506, -0.800989],
                     [-0.787367,  0.773506,  0.800989],
                     [ 0.787367, -0.800989,  0.773506],
                     [ 0.787367,  0.800989, -0.773506],
                     [ 0.814652, -0.80031,  -0.828745],
                     [ 0.814652,  0.80031,   0.828745],
                     [-0.814652, -0.828745,  0.80031 ],
                     [-0.814652,  0.828745, -0.80031]])
        tstress1 = np.array([2861.0152, 2861.0152, 2861.0152, 0, 0, 0])
        
        tenergy2 = 3.9436037 / len(self.struc)
        tforces2 = np.array(
                    [[-0.781195, -0.767443, -0.794710],
                     [-0.781195,  0.767443,  0.794710],
                     [ 0.781195, -0.794710,  0.767443],
                     [ 0.781195,  0.794710, -0.767443],
                     [ 0.808353, -0.794122, -0.822338],
                     [ 0.808353,  0.794122,  0.822338],
                     [-0.808353, -0.822338,  0.794122],
                     [-0.808353,  0.822338, -0.794122]])
        tstress2 = np.array([2838.7443, 2838.7443, 2838.7443, 0, 0, 0])

        energy1 = self.d1['energy'] / len(self.struc)
        forces1 = self.d1['force']
        stress1 = self.d1['stress'] * 1602176.6208
        energy2 = self.d2['energy'] / len(self.struc)
        forces2 = self.d2['force']
        stress2 = self.d2['stress'] * 1602176.6208

        self.assertTrue(abs(tenergy1-energy1) < 1e-7)
        self.assertTrue(np.allclose(tforces1, forces1))
        self.assertTrue(np.allclose(tstress1, stress1))

        self.assertTrue(abs(tenergy2-energy2) < 1e-7)
        self.assertTrue(np.allclose(tforces2, forces2))
        self.assertTrue(np.allclose(tstress2, stress2))

class TestRegression(unittest.TestCase):
    model = {'system' : system,
             'hiddenlayers': [12, 12],
             'epoch': 10,
             'stress_coefficient': None,
             'force_coefficient': 0.03,
             'path': 'unittest/'
            }
    ff = PyXtal_FF(descriptors=descriptor, model=model)
    struc = bulk('Si', 'diamond', a=5.0, cubic=True)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('unittest')
        
    def test_1_NN_LBFGS(self):
        self.ff.algorithm = 'NN'
        self.ff._model['optimizer'] = {'method': 'lbfgs'}
        (train_stat, _) = self.ff.run(mode='train', TrainData=TrainData)

    def test_2_NN_ADAM(self):
        self.ff.algorithm = 'NN'
        self.ff._model['optimizer']= {'method': 'ADAM'}
        self.ff._MODEL(self.ff._model)
        train_stat = self.ff.model.train('Train_db', self.ff.optimizer)
        self.ff.model.save_checkpoint(des_info=self.ff._descriptors)


    def test_3_lr(self):
        self.ff.algorithm = 'PR'
        self.ff._model['order'] = 1
        self.ff._MODEL(self.ff._model)
        train_stat = self.ff.model.train('Train_db', None)
        self.ff.model.save_checkpoint(des_info=self.ff._descriptors)

    def test_4_qr(self):
        self.ff.algorithm = 'PR'
        self.ff._model['order'] = 2
        self.ff._MODEL(self.ff._model)
        train_stat = self.ff.model.train('Train_db', None)

    def test_5_NN_calculator(self):
        #calc = PyXtalFFCalculator(mliap='unittest/12-12-checkpoint.pth', logo=False)
        ff = PyXtal_FF(model={'system': ["Si"]}, logo=False)
        ff.run(mode='predict', mliap='unittest/12-12-checkpoint.pth')
        calc = PyXtalFFCalculator(ff=ff)

        self.struc.set_calculator(calc)
        self.struc.get_potential_energy()
        self.struc.get_stress()

    def test_6_LR_calculator(self):
        #calc = PyXtalFFCalculator(mliap='unittest/PolyReg-checkpoint.pth', logo=False)
        ff = PyXtal_FF(model={'system': ["Si"]}, logo=False)
        ff.run(mode='predict', mliap='unittest/PolyReg-checkpoint.pth')
        calc = PyXtalFFCalculator(ff=ff)

        self.struc.set_calculator(calc)
        self.struc.get_potential_energy()
        self.struc.get_stress()

class TestRegressionComp(unittest.TestCase):

    model = {'system' : system,
             'stress_coefficient': None,
             'force_coefficient': 0.03,
             'path': 'unittest_comp/',
             'algorithm': 'PR',
            }
    ff = PyXtal_FF(descriptors=descriptor_comp, model=model)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree('unittest_comp')
        
    def test_lr_comp(self):
        self.ff._model['order'] = 1
        (train_stat, _) = self.ff.run(mode='train', TrainData=TrainData)

if __name__ == '__main__':

    unittest.main()



import os
import gc
import time
import shelve
import numpy as np
from ase import Atoms, units
from copy import deepcopy
from functools import partial
from torch.utils.data import Dataset
from monty.serialization import loadfn
from multiprocessing import Pool, cpu_count
from collections.abc import MutableSequence
from .base_potential import ZBL


class Database():#MutableSequence):
    def __init__(self, name):
        self.name = name
        self.database = shelve.open(self.name)
        
        self.msg = "Index must be an integer in the interval [0,len(self)]"
        self.length = len(list(self.database.keys()))


    def __len__(self):
        return self.length


    def __setitem__(self, index, value):
        if isinstance(index, int) and index >= 0:
            self.database[str(index)] = value
        else:
            raise IndexError(self.msg)


    def __getitem__(self, index):
        if isinstance(index, int) and index >= 0 and index < len(self):
            return self.database[str(index)]
        else:
            raise IndexError(self.msg)


    def __delitem__(self, index):
        if isinstance(index, int) and index >= 0:
            del(self.database[str(index)])
        else:
            raise IndexError(self.msg)


    def insert(self, index, value):
        """ Insert the value to the dictionary at index. """
        if isinstance(index, int) and index >= 0:
            pass
        else:
            raise IndexError(self.msg)
        self[index] = value


    def append(self, value):
        """ Append value to the end of the sequence. """
        self.insert(len(self), value)


    def close(self):
        self.database.close()


    def store(self, structure_file, function, storage, ase_db=None):
        """ Map structures to descriptors and store them, including features, to database.
        If compute is False, print pre-computed descriptors message. """
        if function['base_potential']:
            self.base_potential = ZBL(function['base_potential']['inner'],
                                      function['base_potential']['outer'])
        else:
            self.base_potential = None

        if storage:
            if os.path.isdir(structure_file) or structure_file.find('json') > 0:
                fmt = 'json'
            elif structure_file.find('xyz') > 0:
                fmt = 'xyz'
            elif structure_file.find('db') > 0:
                fmt = 'db'
            elif structure_file.find('traj') > 0:
                fmt = 'traj'
            else:
                fmt = 'vasp-out'
        
            # extract the structures and energy, forces, and stress information.
            if fmt == 'json':
                data = parse_json(structure_file)
            elif fmt == 'vasp-out':
                data = parse_OUTCAR_comp(structure_file)
            elif fmt == 'xyz':
                data = parse_xyz(structure_file)
            elif fmt == 'db':
                data = parse_ase_db(structure_file)
            elif fmt == 'traj':
                data = parse_traj(structure_file)
            else:
                raise NotImplementedError('PyXtal_FF supports only json, vasp-out, and xyz formats')
            print("{:d} structures have been loaded.".format(len(data)))
            
            self.add(function, data)

            if ase_db is not None and fmt != 'db':
                convert_to_ase_db(data, ase_db)
                print("save the structures to {:}.".format(ase_db))

        else:
            print(f"Features and precomputed descriptors exist: {self.name}.dat\n")


    def add(self, function, data):
        """ Add descriptors for all structures to database. """
        print('Computing the descriptors...')

        _N = deepcopy(function['N'])
        _cpu = deepcopy(function['ncpu'])
        
        N1 = len(data)
        if _N is not None and _N < N1:
            lists = range(_N)
        else:
            lists = range(N1)
            
        if _cpu == 1:
            for i, index in enumerate(lists):
                d = self.compute(function, data[index])
                self.append(d)
                self.length += 1
                print('\r{:4d} out of {:4d}'.format(i+1, len(lists)), flush=True, end='')

        else:
            # we do it for the reduced lists _data
            _data = [data[item] for item in lists]
            failed_ids = []
            with Pool(_cpu) as p:
                func = partial(self.compute, function)
                for i, d in enumerate(p.imap_unordered(func, _data)):
                    try:
                        self.append(d)
                        self.length += 1
                    except: # track the failed structure ids
                        failed_ids.append(i)
                        #pass
                    print('\r{:4d} out of {:4d}'.format(i+1, len(_data)), flush=True, end='')
                p.close()
                p.join()

            # compute the missing structures in parallel calcs
            for id in failed_ids: 
                d = self.compute(function, _data[id])
                self.append(d)
                self.length += 1

        print(f"\nSaving descriptor-feature data to {self.name}.dat\n")

  
    def compute(self, function, data):
        """ Compute descriptor for one structure to the database. """

        if function['type'] in ['BehlerParrinello', 'ACSF']:
            from pyxtal_ff.descriptors.ACSF import ACSF
            d = ACSF(function['parameters'],
                     function['Rc'], 
                     function['force'],
                     function['stress'], 
                     function['cutoff'], False).calculate(data['structure'])

        elif function['type'] in ['wACSF', 'wacsf']:
            from pyxtal_ff.descriptors.ACSF import ACSF
            d = ACSF(function['parameters'],
                     function['Rc'], 
                     function['force'],
                     function['stress'], 
                     function['cutoff'], True).calculate(data['structure'])
        
        elif function['type'] in ['SO4', 'Bispectrum', 'bispectrum']:
            from pyxtal_ff.descriptors.SO4 import SO4_Bispectrum
            d = SO4_Bispectrum(function['parameters']['lmax'],
                               function['Rc'],
                               derivative=function['force'],
                               stress=function['stress'],
                               normalize_U=function['parameters']['normalize_U'],
                               cutoff_function=function['cutoff']).calculate(data['structure'])
        
        elif function['type'] in ['SO3', 'SOAP', 'soap']:
            from pyxtal_ff.descriptors.SO3 import SO3
            d = SO3(function['parameters']['nmax'],
                    function['parameters']['lmax'],
                    function['Rc'],
                    alpha=function['parameters']['alpha'],
                    derivative=function['force'],
                    stress=function['stress']).calculate(data['structure'])

        elif function['type'] in ['EAD', 'ead']:
            from pyxtal_ff.descriptors.EAD import EAD
            d = EAD(function['parameters'],
                     function['Rc'],
                     function['force'], function['stress'],
                     function['cutoff']).calculate(data['structure'])
        
        elif function['type'] in ['SNAP', 'snap']:
            from pyxtal_ff.descriptors.SNAP import SO4_Bispectrum
            d = SO4_Bispectrum(function['weights'],
                               function['parameters']['lmax'],
                               function['Rc'],
                               derivative=function['force'],
                               stress=function['stress'],
                               normalize_U=function['parameters']['normalize_U'],
                               cutoff_function=function['cutoff'],
                               rfac0=function['parameters']['rfac']).calculate(data['structure'])

        else:
            msg = f"{function['type']} is not implemented"
            raise NotImplementedError(msg)

        if d['rdxdr'] is not None:
            N = d['x'].shape[0]
            L = d['x'].shape[1]
            rdxdr = np.zeros([N, L, 3, 3])
            for _m in range(N):
                ids = np.where(d['seq'][:,0]==_m)[0]
                rdxdr[_m, :, :, :] += np.einsum('ijkl->jkl', d['rdxdr'][ids, :, :, :])
            d['rdxdr'] = rdxdr.reshape([N, L, 9])[:, :, [0, 4, 8, 1, 2, 5]]
            #d['rdxdr'] = np.einsum('ijklm->iklm', d['rdxdr'])\
            #.reshape([shp[0], shp[2], shp[3]*shp[4]])[:, :, [0, 4, 8, 1, 2, 5]]  #need to change
        
        #print(len(data['structure']))
        if self.base_potential:
            base_d = self.base_potential.calculate(data['structure'])
        else:
            base_d = {'energy': 0., 'force': 0., 'stress': 0.}
        
        #print(data['energy'])
        #print(base_d['energy'])

        d['energy'] = np.asarray(data['energy'] - base_d['energy'])
        d['force'] = np.asarray(data['force']) - base_d['force']
        if data['stress'] is not None:
            d['stress'] = np.asarray(data['stress']) - base_d['stress'] / units.GPa
        else:
            d['stress'] = data['stress'] 
        d['group'] = data['group']

        return d


def compute_descriptor(function, structure):
    """ Compute descriptor for one structure. """

    if function['type'] in ['BehlerParrinello', 'ACSF']:
        from pyxtal_ff.descriptors.ACSF import ACSF
        d = ACSF(function['parameters'],
                 function['Rc'], 
                 function['force'],
                 function['stress'],
                 function['cutoff'], False).calculate(structure)

    elif function['type'] in ['wACSF', 'wacsf']:
        from pyxtal_ff.descriptors.ACSF import ACSF
        d = ACSF(function['parameters'],
                 function['Rc'], 
                 function['force'],
                 function['stress'],
                 function['cutoff'], True).calculate(structure)

    elif function['type'] in ['SO4', 'Bispectrum', 'bispectrum']:
        from pyxtal_ff.descriptors.SO4 import SO4_Bispectrum
        d = SO4_Bispectrum(function['parameters']['lmax'],
                           function['Rc'],
                           derivative=True,
                           stress=True,
                           normalize_U=function['parameters']['normalize_U'],
                           cutoff_function=function['cutoff']).calculate(structure)

    elif function['type'] in ['SO3', 'SOAP', 'soap']:
        from pyxtal_ff.descriptors.SO3 import SO3
        d = SO3(function['parameters']['nmax'],
                function['parameters']['lmax'],
                function['Rc'],
                derivative=True,
                stress=True).calculate(structure)

    elif function['type'] in ['EAD', 'ead']:
            from pyxtal_ff.descriptors.EAD import EAD
            d = EAD(function['parameters'],
                     function['Rc'],
                     True, True,
                     function['cutoff']).calculate(structure)

    elif function['type'] in ['SNAP', 'snap']:
        from pyxtal_ff.descriptors.SNAP import SO4_Bispectrum
        d = SO4_Bispectrum(function['weights'],
                           function['parameters']['lmax'],
                           function['Rc'],
                           derivative=True,
                           stress=True,
                           normalize_U=function['parameters']['normalize_U'],
                           cutoff_function=function['cutoff']).calculate(structure)

    else:
        msg = f"{function['type']} is not implemented"
        raise NotImplementedError(msg)
    
    if d['rdxdr'] is not None:
        N = d['x'].shape[0]
        L = d['x'].shape[1]
        rdxdr = np.zeros([N, L, 3, 3])
        for _m in range(N):
            ids = np.where(d['seq'][:,0]==_m)[0]
            rdxdr[_m, :, :, :] += np.einsum('ijkl->jkl', d['rdxdr'][ids, :, :, :])
        d['rdxdr'] = rdxdr.reshape([N, L, 9])[:, :, [0, 4, 8, 1, 2, 5]]
 
    return d


def parse_json(path, N=None, Random=False):
    """ Extract structures/energy/forces/stress information from json file. """
    if os.path.isfile(path):
        structure_dict = loadfn(path)
    elif os.path.isdir(path):
        import glob
        cwd = os.getcwd()
        os.chdir(path)
        files = glob.glob('*.json')
        os.chdir(cwd)
        structure_dict = []
        for file in files:
            fp = os.path.join(path, file)
            structure_dict += loadfn(fp)

    if N is None:
        N = len(structure_dict)
    elif Random and N < len(structure_dict):
        structure_dict = sample(structure_dict, N)

    data = []
    for i, d in enumerate(structure_dict):
        if 'structure' in d:
            structure = Atoms(symbols=d['structure'].atomic_numbers,
                              positions=d['structure'].cart_coords,
                              cell=d['structure'].lattice._matrix, pbc=True)
            v = structure.get_volume()
            if 'data' in d:
                key = 'data'
            else:
                key = 'outputs'
            
            if 'energy_per_atom' in d[key]:
                energy = d[key]['energy_per_atom']*len(structure)
            else:
                energy = d[key]['energy']
            force = d[key]['forces']
            try:
                if d['tags'][0] == 'Strain':
                    group = 'Elastic'
                else:
                    group = 'NoElastic'
            except:
                if d['group'] == 'Elastic':
                    group = 'Elastic'
                else:
                    group = 'NoElastic'
            #group = d['group']
            # vasp default output: XX YY ZZ XY YZ ZX
            # pyxtal_ff/lammps use: XX YY ZZ XY XZ YZ
            # Here we assume the sequence is lammps
            if d['group'] == 'Elastic' and 'Mo 3x3x3 cell' in d['description']:
                stress = None
                group = 'NoElastic'
            elif 'virial_stress' in d[key]: #kB to GPa
                s = [s/10 for s in d[key]['virial_stress']] 
                if d['group'] == 'Ni3Mo' or d['element'] == 'Cu': #just tentative fix
                    stress = [s[0], s[1], s[2], s[3], s[4], s[5]]
                else:
                    stress = [s[0], s[1], s[2], s[3], s[5], s[4]]

            elif 'stress' in d[key]: #kB to GPa
                s = [s/10 for s in d[key]['stress']]
                if d['group'] == 'Ni3Mo': #to fix the issue
                    stress = [s[0], s[1], s[2], s[3], s[4], s[5]]
                else:
                    stress = [s[0], s[1], s[2], s[3], s[5], s[4]]
            else:
                stress = None
            
            data.append({'structure': structure,
                         'energy': energy, 'force': force, 
                         'stress': stress, 'group': group})
        
        else:   # For PyXtal
            structure = Atoms(symbols=d['elements'], scaled_positions=d['coords'], 
                              cell=d['lattice'], pbc=True)
            data.append({'structure': structure,
                         'energy': d['energy'], 'force': d['force'], 
                         'stress': None, 'group': 'random'})
           
        if i == (N-1):
            break


    return data


def create_label(elements, hiddenlayers):
    label = ''
    for e in elements:
        label += e
        label += '-'
    for l in hiddenlayers:
        label += str(l)
        label += '-'
    label += '1'
    return label


def get_descriptors_parameters(symmetry, system):
    from itertools import combinations_with_replacement
    G = []
    if 'G2' in symmetry:
        combo = list(combinations_with_replacement(system, 1))
        if 'Rs' not in symmetry['G2']:
            Rs = [0.]
        else:
            Rs = symmetry['G2']['Rs']
        
        for eta in symmetry['G2']['eta']:
            for rs in Rs:
                for element in combo:
                    g = [element[0], 'nan', rs, eta, 'nan', 'nan']
                    G.append(g)

    if 'G4' in symmetry:
        combo = list(combinations_with_replacement(system, 2))
        for zeta in symmetry['G4']['zeta']:
            for lamBda in symmetry['G4']['lambda']:
                for eta in symmetry['G4']['eta']:
                    for p_ele in combo:
                        g = [p_ele[0], p_ele[1], 'nan', eta, lamBda, zeta]
                        G.append(g)

    return G


def parse_xyz(structure_file, N=1000000):
    """
    Extract structures/enegy/force information of xyz file provided by the Cambridge group
    """
    data = []
    with open(structure_file, 'r') as f:
        lines = f.readlines()
        count = 0
        while True:
            line = lines[count]
            symbols = []
            number = int(line)
            coords = np.zeros([number, 3])
            forces = np.zeros([number, 3])
            infos = lines[count+1].split('=')
            for i, info in enumerate(infos):
                if info.find('dft_energy') >= 0 or info.find('DFT_energy') >= 0:
                    energy = float(infos[i+1].split()[0])
                elif info.find('config_type') == 0:
                    group = infos[i+1].split()[0]
                elif info.find('Lattice') == 8:
                    lat = []
                    tmp = infos[i+1].split()
                    for num in tmp:
                        num = num.replace('"','')
                        if num.replace('.', '', 1).replace('-', '', 1).isdigit():
                            lat.append(float(num))
                    lat = np.array(lat).reshape([3,3])
                elif info.find('virial') > 0:
                    s = []
                    tmp = infos[i+1].split()
                    for num in tmp:
                        num = num.replace('"','')
                        if num.replace('.', '', 1).replace('-', '', 1).isdigit():
                            s.append(float(num))
                    s = np.array(s).flatten()
                    stress = np.array([s[0], s[4], s[8], s[1], s[2], s[5]])
                elif info.find('Properties') >= 0: 
                    items = infos[i+1].split(':')
                    counts = 0
                    for i in range(int(len(items)/3)):
                        if items[i*3] == 'dft_force' or items[i*3] == 'DFT_force':
                            count_f = counts
                        elif items[i*3] == 'pos':
                            count_p = counts
                        counts += int(items[i*3+2])
                    break

            for i in range(number):
                infos = lines[count+2+i].split()
                symbols.append(infos[0])
                coords[i, :] = np.array([float(num) for num in infos[count_p:count_p+3]])
                forces[i, :] = np.array([float(num) for num in infos[count_f:count_f+3]])

            structure = Atoms(symbols=symbols,
                              positions=coords,
                              cell=lat, pbc=True)
            data.append({'structure': structure,
                         'energy': energy, 'force': forces,
                         'stress': stress, 'group': group})

            count = count + number + 2
            if count >= len(lines) or len(data) == N:
                break

    return data

def parse_OUTCAR_comp(structure_file, N=1000000):
    """
    Extract structures/enegy/force information from compressed OUTCAR file
    """
    data = []
    with open(structure_file, 'r') as f:
        lines = f.readlines()
        read_symbol = True
        read_composition = False
        read_structure = False
        symbols = []
        count = 0
        while True:
            line = lines[count]
            if read_symbol:
                if line.find('POTCAR') > 0:
                    symbol = line.split()[2]
                    if symbol in symbols:
                        read_symbol = False
                        read_composition = True
                    else:
                        symbols.append(symbol)
            elif read_composition:
                if line.find('ions per type') > 0:
                    symbol_array = []
                    tmp = line.split('=')[1].split()
                    numIons = []
                    for i in range(len(symbols)):
                        for j in range(int(tmp[i])):
                            symbol_array.append(symbols[i])
                    read_composition = False
                    read_structure = True
                    count += 4
                    # search for "   energy  without entropy="
                    line_number = 0
                    for tmp in range(24):
                        if lines[count+len(symbol_array)+tmp].find("entropy") > 0:
                            line_number = len(symbol_array) + tmp
                            break
                    if line_number == 0:
                        raise ValueError("Cannot identify the line starts with energy without entropy")

            else: #read_structure
                lat_string = lines[count+1:count+4]
                lat = np.zeros([3,3])
                for i, l in enumerate(lines[count+1:count+4]):
                    lat[i,:] = np.fromstring(l, dtype=float, sep=' ')[:3]
                coor = np.zeros([len(symbol_array), 3])
                force = np.zeros([len(symbol_array), 3])
                for i, l in enumerate(lines[count+6:count+6+len(symbol_array)]):
                    array = np.fromstring(l, dtype=float, sep=' ')
                    coor[i,:] = array[:3]
                    force[i,:] = array[3:]
                
                energy = float(lines[count+line_number-1].split()[-4])
                structure = Atoms(symbols=symbol_array,
                                  positions=coor,
                                  cell=lat, pbc=True)
                data.append({'structure': structure,
                             'energy': energy, 'force': force,
                             'stress': None, 'group': 'random'})
                count += line_number - 1

            count += 1
            if count >= len(lines) or len(data) == N:
                break

    return data

def convert_to_ase_db(data, db_path='test.db'):
    from ase.db import connect
    
    with connect(db_path) as db:
        for d in data:
            struc = d['structure']
            d.pop('structure', None)
            db.write(struc, data=d)

def parse_ase_db(db_path, N=None, Random=False):
    from ase.db import connect

    data = []
    with connect(db_path) as db:
        if N is None:
            N = len(db)

        for i, row in enumerate(db.select()):
            structure = db.get_atoms(row.id)
            if "stress" in row.data:
                stress = row.data["stress"]
            else:
                stress = None

            if "group" in row.data:
                group = row.data["group"]
            else:
                group = None

            if "dft_energy" in row.data:
                eng = row.data["dft_energy"]
            else:
                eng = row.data["energy"]

            if "dft_force" in row.data:
                force = row.data["dft_force"]
            else:
                force = row.data["force"]
            data.append({'structure': structure,
                         'energy': eng,
                         'force': force,
                         'group': group,
                         'stress': stress})

            if i == (N-1):
                break

    return data

def parse_traj(structure_file):
    from ase.build import sort
    from ase.io.trajectory import Trajectory
    data = []
    out_traj = Trajectory(structure_file)
    for traj in out_traj:
        structure = sort(traj)
        energy = traj.get_potential_energy()
        force = traj.get_forces()
        xjson = {'structure':structure,
                 'energy':energy,
                 'force':force,
                 'stress':None,
                 'group':'random'}
        data.append(xjson)
    return data



"""ASE LAMMPS Calculator Library Version"""
from __future__ import print_function
import numpy as np
import ase.units
from ase.calculators.calculator import Calculator
import os
#

def is_upper_triangular(arr, atol=1e-8):
    """test for upper triangular matrix based on numpy"""
    # must be (n x n) matrix
    assert len(arr.shape) == 2
    assert arr.shape[0] == arr.shape[1]
    return np.allclose(np.tril(arr, k=-1), 0., atol=atol) and \
        np.all(np.diag(arr) >= 0.0)


def convert_cell(ase_cell):
    """
    Convert a parallelepiped (forming right hand basis)
    to lower triangular matrix LAMMPS can accept. This
    function transposes cell matrix so the bases are column vectors
    """
    cell = ase_cell.T
    if not is_upper_triangular(cell):
        tri_mat = np.zeros((3, 3))
        A = cell[:, 0]
        B = cell[:, 1]
        C = cell[:, 2]
        tri_mat[0, 0] = np.linalg.norm(A)
        Ahat = A / np.linalg.norm(A)
        AxBhat = np.cross(A, B) / np.linalg.norm(np.cross(A, B))
        tri_mat[0, 1] = np.dot(B, Ahat)
        tri_mat[1, 1] = np.linalg.norm(np.cross(Ahat, B))
        tri_mat[0, 2] = np.dot(C, Ahat)
        tri_mat[1, 2] = np.dot(C, np.cross(AxBhat, Ahat))
        tri_mat[2, 2] = np.linalg.norm(np.dot(C, AxBhat))
        # create and save the transformation for coordinates
        volume = np.linalg.det(ase_cell)
        trans = np.array([np.cross(B, C), np.cross(C, A), np.cross(A, B)])
        trans /= volume
        coord_transform = np.dot(tri_mat, trans)
        return tri_mat, coord_transform
    else:
        return cell, None


class LAMMPSlib(Calculator):

    def __init__(self, lmp, lmpcmds=None, path='tmp', 
                lmp_file=None, ntyp=None, *args, **kwargs):
        Calculator.__init__(self, *args, **kwargs)
        self.lmp = lmp
        self.lmp_file = lmp_file
        self.folder = path
        self.ntyp = ntyp
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        self.lammps_data = self.folder+'/data.lammps'
        self.lammps_in = self.folder + '/in.lammps'
        self.lmpcmds = lmpcmds
        self.paras = []
        for para in lmpcmds:
            self.paras.append(para)

    def __str__(self):
        s = "\nLAMMPS calculator with pyxtal_ff force field\n"
        return s

    def __repr__(self):
        return str(self)


    def calculate(self, atoms):
        """
        prepare lammps .in file and data file
        write_lammps_data(filename, self.atoms, )
        """
        boundary = ''
        for i in range(3):
            if atoms.pbc[i]:
                boundary += 'p ' 
            else:
                boundary += 'f '
        if boundary in ['f f p ', 'p p f ']: #needs some work later
            boundary = 'p p p '
        self.boundary = boundary
        self.write_lammps_data(atoms)
        self.write_lammps_in()
        self.lmp.file(self.lammps_in)

        # Extract the forces and energy
        self.lmp.command('variable pxx equal pxx')
        self.lmp.command('variable pyy equal pyy')
        self.lmp.command('variable pzz equal pzz')
        self.lmp.command('variable pxy equal pxy')
        self.lmp.command('variable pxz equal pxz')
        self.lmp.command('variable pyz equal pyz')
        self.lmp.command('variable fx atom fx')
        self.lmp.command('variable fy atom fy')
        self.lmp.command('variable fz atom fz')
        self.lmp.command('variable pe equal pe')

        pos = np.array(
                [x for x in self.lmp.gather_atoms("x", 1, 3)]).reshape(-1, 3)
        
        self.energy = self.lmp.extract_variable('pe', None, 0) 

        xlo = self.lmp.extract_global("boxxlo")
        xhi = self.lmp.extract_global("boxxhi")
        ylo = self.lmp.extract_global("boxylo")
        yhi = self.lmp.extract_global("boxyhi")
        zlo = self.lmp.extract_global("boxzlo")
        zhi = self.lmp.extract_global("boxzhi")
        xy = self.lmp.extract_global("xy")
        yz = self.lmp.extract_global("yz")
        xz = self.lmp.extract_global("xz")
        unitcell = np.array([[xhi-xlo, xy,  xz],
                               [0,  yhi-ylo,  yz],
                               [0,   0,  zhi-zlo]]).T

        stress = np.empty(6)
        stress_vars = ['pxx', 'pyy', 'pzz', 'pyz', 'pxz', 'pxy']

        for i, var in enumerate(stress_vars):
            stress[i] = self.lmp.extract_variable(var, None, 0)

        stress_mat = np.zeros((3, 3))
        stress_mat[0, 0] = stress[0]
        stress_mat[1, 1] = stress[1]
        stress_mat[2, 2] = stress[2]
        stress_mat[1, 2] = stress[3]
        stress_mat[2, 1] = stress[3]
        stress_mat[0, 2] = stress[4]
        stress_mat[2, 0] = stress[4]
        stress_mat[0, 1] = stress[5]
        stress_mat[1, 0] = stress[5]
        stress[0] = stress_mat[0, 0]
        stress[1] = stress_mat[1, 1]
        stress[2] = stress_mat[2, 2]
        stress[3] = stress_mat[1, 2]
        stress[4] = stress_mat[0, 2]
        stress[5] = stress_mat[0, 1]

        self.stress = -stress * 1e5 * ase.units.Pascal
        f = (np.array(self.lmp.gather_atoms("f", 1, 3)).reshape(-1,3) *
                (ase.units.eV/ase.units.Angstrom))
        self.forces = f.copy()
        atoms.positions = pos.copy()
        atoms.cell = unitcell.copy()
        self.atoms = atoms.copy()

    def write_lammps_in(self):
        if self.lmp_file is not None:
            os.system('cp ' + self.lmp_file + ' ' + self.folder + '/in.lammps')
        else:
            with open(self.lammps_in, 'w') as fh:
                fh.write('clear\n')
                fh.write('box  tilt large\n')
                fh.write('units metal\n')
                fh.write('boundary {:s}\n'.format(self.boundary))
                fh.write('atom_modify sort 0 0.0\n') 
                fh.write('read_data {:s}\n'.format(self.lammps_data))
                fh.write('\n### interactions\n')
                for para in self.paras:
                    fh.write("{:s}\n".format(para))
                fh.write('neighbor 1.0 bin\n')
                fh.write('neigh_modify  every 1  delay 1  check yes\n')
                fh.write('thermo_style custom pe pxx pyy pzz pyz pxz pxy\n')
                fh.write('thermo_modify flush yes\n')
                fh.write('thermo 1\n')
                fh.write('run 1\n')

    def write_lammps_data(self, atoms):
        # QZ: assign the tags by sorting the element
        # eg [Si, Si, Si, O, O] will be [2, 2, 2, 1, 1]
        # Si, O will be sorted alphabetically
        atom_types = np.array(atoms.get_tags()) #[1]*len(atoms)
        if sum(atom_types) == 0:
            symbols = atoms.get_chemical_symbols()
            ordered_eles = sorted(list(set(symbols)))
            ntypes = len(ordered_eles)
            tags = [] 
            for symbol in symbols:
                for t, e in enumerate(ordered_eles):
                    if e == symbol:
                        tags.append(t+1)
                        break
        else:
            ntypes = len(np.unique(atom_types))
            tags = atom_types 
        
        # force n types
        if self.ntyp is not None:
            ntypes = self.ntyp

        with open(self.lammps_data, 'w') as fh:
            comment = 'lammpslib autogenerated data file'
            fh.write(comment.strip() + '\n\n')
            fh.write('{:d} atoms\n'.format(len(atoms)))
            fh.write('{:d} atom types\n'.format(ntypes))
            cell, coord_transform = convert_cell(atoms.get_cell())
            fh.write('\n')
            fh.write('{0:16.8e} {1:16.8e} xlo xhi\n'.format(0.0, cell[0, 0]))
            fh.write('{0:16.8e} {1:16.8e} ylo yhi\n'.format(0.0, cell[1, 1]))
            fh.write('{0:16.8e} {1:16.8e} zlo zhi\n'.format(0.0, cell[2, 2]))
            fh.write('{0:16.8e} {1:16.8e} {2:16.8e} xy xz yz\n'
                             ''.format(cell[0, 1], cell[0, 2], cell[1, 2]))

            fh.write('\n\nAtoms \n\n')
            for i, (typ, pos) in enumerate(
                    zip(tags, atoms.get_positions())):
                if coord_transform is not None:
                    pos = np.dot(coord_transform, pos.transpose())
                fh.write('{:4d} {:4d} {:16.8e} {:16.8e} {:16.8e}\n'
                         .format(i + 1, typ, pos[0], pos[1], pos[2]))

    def update(self, atoms):
        if not hasattr(self, 'atoms') or self.atoms != atoms:
            self.calculate(atoms)

    def get_potential_energy(self, atoms, force_consistent=False):
        self.update(atoms)
        return self.energy

    def get_forces(self, atoms):
        self.update(atoms)
        return self.forces.copy()

    def get_stress(self, atoms):
        self.update(atoms)
        return self.stress.copy()


if __name__ == '__main__':
    from ase.io import read
    from ase.build import bulk
    from lammps import lammps

    lammps_name=''
    comm=None
    log_file='lammps.log'
    cmd_args = ['-echo', 'log', '-log', log_file,
                '-screen', 'none', '-nocite']
    lmp = lammps(lammps_name, cmd_args, comm)

    struc = bulk('Si', 'diamond', cubic=True)
    struc.set_tags([1]*len(struc))
    parameters = ["mass * 1.0",
                  "pair_style mliap model nn Si-snap/NN_weights.txt descriptor sna Si-snap/DescriptorParams.txt",
                  "pair_coeff * * Si Si"
                  ]

    lammps = LAMMPSlib(lmp=lmp, lmpcmds=parameters)
    struc.set_calculator(lammps)
    print('positions: ')
    print(struc.get_positions())
    print('energy: ')
    print(struc.get_potential_energy())
    print('force: ')
    print(struc.get_forces())
    print('stress: ')
    print(struc.get_stress())




#!/usr/bin/env  python
# encoding: utf-8

class Regressor:
    """ This class defines the local optimization method.

    Parameters
    ----------
    method: str
        Type of minimization scheme, e.g.: 'LBFGS'.
    user_kwargs: dict
        The arguments for the optimization method. These arguments are 
        passed with a dict.
    """
    def __init__(self, method, user_kwargs):
        self.method = method
        
        kwargs = {'lr': 1,}
        if method in ['lbfgs', 'LBFGS']:
            from pyxtal_ff.models.optimizers.lbfgs import LBFGS as optimizer
            _kwargs = {'max_iter': 100,
                       'max_eval': 15000,
                       'tolerance_grad': 1e-7,
                       'tolerance_change': 1e-9,
                       'history_size': 10,
                       'line_search_fn': 'strong_wolfe'}
        
        elif self.method in ['SGD', 'sgd']:
            from torch.optim import SGD as optimizer
            _kwargs = {'lr': 0.001,
                       'momentum': 0.,
                       'dampening': 0.,
                       'weight_decay': 0,
                       'nesterov': False}
        
        elif self.method in ['Adam', 'ADAM', 'adam']:
            from torch.optim import Adam as optimizer
            _kwargs = {'lr': 0.001,
                       'betas': (0.9, 0.999), 
                       'eps': 1e-08,
                       'weight_decay': 0, 
                       'amsgrad': False}

        elif self.method in ['lbfgsb']:
            from pyxtal_ff.models.optimizers.lbfgsb import LBFGSScipy as optimizer
            _kwargs = {'max_iter': 100,
                       'max_eval': 15000,
                       'tolerance_grad': 1e-7,
                       'tolerance_change': 1e-9,
                       'history_size': 10}

        else:
            msg = f"The {method} is not implemented yet."
            raise NotImplementedError(msg)
            
        kwargs.update(_kwargs)
        
        if user_kwargs is not None:
            kwargs.update(user_kwargs)

        self.optimizer = optimizer
        self.kwargs = kwargs


    def regress(self, models):
        """ Define optimization scheme and return the optimizer to models.

        Parameters
        ----------
        models: object
            Class representing the regression model.

        Returns
        -------
        regressor
            PyTorch optimizer or Scipy LBFGS-B.
        """
        try:
            params = models['model'].parameters()
        except:
            params = [p for model in models.values() for p in model.parameters()]

        if self.method in ['LBFGS', 'lbfgs']:
            regressor = self.optimizer(params, 
                                       lr=self.kwargs['lr'],
                                       max_iter=self.kwargs['max_iter'],
                                       max_eval=self.kwargs['max_eval'],
                                       tolerance_grad=self.kwargs['tolerance_grad'],
                                       tolerance_change=self.kwargs['tolerance_change'],
                                       history_size=self.kwargs['history_size'],
                                       line_search_fn=self.kwargs['line_search_fn'])

        elif self.method in ['SGD', 'sgd']:
            regressor = self.optimizer(params,
                                       lr=self.kwargs['lr'],
                                       momentum=self.kwargs['momentum'],
                                       dampening=self.kwargs['dampening'],
                                       weight_decay=self.kwargs['weight_decay'],
                                       nesterov=self.kwargs['nesterov'])

        elif self.method in ['adam', 'ADAM', 'Adam']:
            regressor = self.optimizer(params,
                                       lr=self.kwargs['lr'],
                                       betas=self.kwargs['betas'],
                                       eps=self.kwargs['eps'],
                                       weight_decay=self.kwargs['weight_decay'],
                                       amsgrad=self.kwargs['amsgrad'])

        elif self.method in ['lbfgsb']: 
            regressor = self.optimizer(params,
                                       max_iter=self.kwargs['max_iter'],
                                       max_eval=self.kwargs['max_eval'],
                                       tolerance_grad=self.kwargs['tolerance_grad'],
                                       tolerance_change=self.kwargs['tolerance_change'],
                                       history_size=self.kwargs['history_size'])
       
        return regressor



import numpy as np
from ase.neighborlist import NeighborList

factor = 14.399645306701126

class ZBL:
    """A class for calculating the Ziegler-Biersack-Littmark (ZBL) 
    screened nuclear repulsion for describing high-energy collisions between atoms.

    Source: J.F. Ziegler, J. P. Biersack and U. Littmark, 
            “The Stopping and Range of Ions in Matter,” Volume 1, Pergamon, 1985.

    Parameters
    ----------
    inner: float
        distance where switching function begins.
    outer: float
        global cutoff for ZBL interaction.
    atomic_energy: bool
        If True, atomic_energy will be recorded and returned.
    """
    def __init__(self, inner, outer, atomic_energy=False):
        self.inner = inner
        self.outer = outer
        self.atomic_energy = atomic_energy

    def calculate(self, crystal):
        """ The calculate function.
        
        Parameters
        ----------
        crystal: object
            ASE Structure object.

        Returns
        -------
        results: dict
            The energy, forces, and stress of ZBL contribution.
        """
        self.crystal = crystal
        self.total_atoms = len(crystal)
        vol = crystal.get_volume()

        rc = [(2.0+self.outer)/2.] * self.total_atoms
        neighbors = NeighborList(rc, self_interaction=False, bothways=False, skin=0.)
        #neighbors = NeighborList(rc, self_interaction=False, bothways=True, skin=0)
        neighbors.update(crystal)

        self.result = {'energy': 0, 'force': np.zeros([self.total_atoms, 3]),
                       'stress': np.zeros([3,3]),
                       'energies': None}
        if self.atomic_energy:
            self.result['energies'] = []

        # Get unique Zi and Zj
        elements = []
        for i in range(self.total_atoms):
            if crystal[i].number not in elements:
                elements.append(crystal[i].number)

        # Get A, B, and C coefficients
        ABC = {}
        for i in range(len(elements)):
            abc = get_ABC_coefficients(elements[i], elements[i], self.outer, self.inner)
            ABC[(elements[i], elements[i])] = abc
            for j in range(i+1, len(elements)):
                abc = get_ABC_coefficients(elements[i], elements[j], self.outer, self.inner)
                ABC[(elements[i], elements[j])] = abc
                ABC[(elements[j], elements[i])] = abc

        for i in range(self.total_atoms):
            element = crystal.get_chemical_symbols()[i]
            elementi = crystal[i].number
            indices, offsets = neighbors.get_neighbors(i)
            Zi, Zj, ABCij = crystal[i].number, [], []
            
            if len(indices) > 0:
                Ri = crystal.get_positions()[i]
                total_neighbors = len(indices)

                Rj = np.zeros([total_neighbors, 3])
                IDs = np.zeros(total_neighbors, dtype=int)

                count = 0
                for j, offset in zip(indices, offsets):
                    Rj[count, :] = crystal.positions[j] + np.dot(offset, crystal.get_cell())
                    IDs[count] = j
                    elementj = crystal[j].number
                    Zj.append(elementj)
                    ABCij.append(ABC[(elementi, elementj)])
                    count += 1
                Zj = np.array(Zj)
                ABCij = np.array(ABCij)

                Rij = Rj - Ri
                Dij = np.sqrt(np.sum(Rij**2, axis=1))

                energy, forces, stress = calculate_ZBL(i, Ri, Rij, Dij, Zi, Zj, self.outer, self.inner, ABCij, self.total_atoms, IDs)
                self.result['energy'] += energy
                self.result['force'] += forces
                self.result['stress'] += stress
                if self.atomic_energy:
                    self.result['energies'].append(energy)
            else:
                if self.atomic_energy:
                    self.result['energies'].append(energy)
        
        if self.atomic_energy:
            self.result['_energies'] = np.array(self.result['energies'])
        self.result['stress'] /= vol
        self.result['stress'] = self.result['stress'].ravel()[[0,4,8,1,2,5]]

        return self.result


def calculate_ZBL(i, ri, rij, dij, Zi, Zj, r_outer, r_inner, ABC, total_atoms, IDs, derivative=True):
    """Calculate the atomic ZBL energy, force, and stress.

    Parameters
    ----------
    i: int
        The i-th center atom
    rij: float array [j,3]
        The vector distances of atom i to neighbors js.
    dij: float array [j]
        The distance between atom i and neighbors js.
    Zi: int
        The atomic number of atom i.
    Zj: int array
        The atomic number of atom j.
    r_outer: float
        global cutoff for ZBL interaction.
    r_inner: float
        distance where switching function begins.
    total_atoms: int
        The total atom in the unit cell
    IDs: int array [j]
        The indices of neighbors centering about atom i.
    """
    ids1 = (dij <= r_outer)

    if True not in ids1:
        return 0., np.zeros([total_atoms, 3]), np.zeros([3,3])
    else:
        _i = i 
        ij_list = i * np.ones([len(IDs), 2], dtype=int)
        ij_list[:, 1] = IDs

        energy = np.zeros([len(dij)])
        forces = np.zeros([total_atoms, 3])
        stress = np.zeros([total_atoms, len(ij_list), 3, 3])
        
        kZi = factor * Zi
        kZiZj = kZi * Zj
        dij_inv = 1 / dij
        Zi_inv = Zi ** 0.23 / 0.46850
        Zj_inv = Zj ** 0.23 / 0.46850
        a_inv = Zi_inv + Zj_inv
        x = dij * a_inv
        
        exp1 = np.exp(-3.19980 * x)
        exp2 = np.exp(-0.94229 * x)
        exp3 = np.exp(-0.40290 * x)
        exp4 = np.exp(-0.20162 * x)
        phi_ij = 0.18175 * exp1 + 0.50986 * exp2 + 0.28022 * exp3 + 0.02817 * exp4
        Eij = kZiZj * dij_inv * phi_ij 
        
        # Switching function
        if derivative:
            dSA = ABC[:,0] * (dij - r_inner) ** 2
            dSB = ABC[:,1] * (dij - r_inner) ** 3
            SA = 0.333333333 * dSA * (dij - r_inner)
            SB = 0.25 * dSB * (dij - r_inner)
            SC = ABC[:,2]
        else:
            SA = 0.333333333 * ABC[:,0] * (dij - r_inner) ** 3
            SB = 0.25 * ABC[:,1] * (dij - r_inner) ** 4
            SC = ABC[:,2]
                
        # Collecting atomic energy
        energy[ids1] += SC[ids1] + Eij[ids1]
        #ids2 = (dij <= r_outer) & (dij > r_inner)
        #ids3 = (dij <= r_outer)
        ids2 = (dij < r_outer) & (dij > r_inner)
        ids3 = (dij < r_outer)
        energy[ids2] += SA[ids2] + SB[ids2]

        # Force
        dE1_ddij = -Eij * dij_inv
        dphi_ddij = (-0.18175 * 3.19980 * exp1 - 0.50986 * 0.94229 * exp2 - \
                      0.28022 * 0.40290 * exp3 - 0.02817 * 0.20162 * exp4) * a_inv
        dE2_ddij = kZiZj * dij_inv * dphi_ddij
        dE_ddij = dE1_ddij + dE2_ddij
        dE_ddij[ids2] += dSA[ids2] + dSB[ids2]

        dRij_dRm = np.zeros([len(dij), 3, total_atoms])
        i_dRij_dRm = np.zeros([len(dij), 3, total_atoms])
        j_dRij_dRm = np.zeros([len(dij), 3, total_atoms])
        for mm in range(total_atoms):
            mm_list = mm * np.ones([len(dij), 1], dtype=int)
            dRij_dRm[:,:,mm], i_dRij_dRm[:,:,mm], j_dRij_dRm[:,:,mm] = dRij_dRm_norm(rij, np.hstack((ij_list, mm_list))) # [j,3,N]

        forces -= np.einsum('ijk,i->kj', dRij_dRm[ids3], dE_ddij[ids3])
        i_force = np.einsum('ijk,i->ikj', i_dRij_dRm, dE_ddij) # [j, N, 3]
        j_force = np.einsum('ijk,i->ikj', j_dRij_dRm, dE_ddij) # [j, N, 3]
        
        # Stress
        for count, ij in enumerate(ij_list):
            _j = ij[1]
            stress[_i, count, :, :] += np.einsum('i,j->ij', i_force[count,_i,:], ri)
            stress[_j, count, :, :] += np.einsum('i,j->ij', j_force[count,_j,:], rij[count]+ri)
        
        stress = -1 * np.einsum('ijkl->kl', stress[:, (dij <= r_outer)])
            
        return np.sum(energy), forces, stress


def get_ABC_coefficients(Zi, Zj, r_outer, r_inner):
    """A function to get the switching A, B, and C coefficients."""
    kZiZj = factor * Zi * Zj
    a_inv = (Zi ** 0.23 + Zj ** 0.23) / 0.46850
    r = r_outer - r_inner
    x = r_outer * a_inv

    exp1 = np.exp(-3.19980 * x)
    exp2 = np.exp(-0.94229 * x)
    exp3 = np.exp(-0.40290 * x)
    exp4 = np.exp(-0.20162 * x)
    
    phi = 0.18175 * exp1 + 0.50986 * exp2 + 0.28022 * exp3 + 0.02817 * exp4
    phiP = (-0.18175 * 3.19980 * exp1 - 0.50986 * 0.94229 * exp2 - \
             0.28022 * 0.40290 * exp3 - 0.02817 * 0.20162 * exp4) * a_inv
    phiDP = (0.18175 * 3.19980 ** 2 * exp1 + 0.50986 * 0.94229 ** 2 * exp2 + \
             0.28022 * 0.40290 ** 2 * exp3 + 0.02817 * 0.20162 ** 2 * exp4) * (a_inv ** 2)

    E = kZiZj * (1 / r_outer) * phi
    EP = kZiZj * (-1 / r_outer ** 2 * phi + 1 / r_outer * phiP)
    EDP = kZiZj * (2 / r_outer ** 3 * phi - 2 / r_outer ** 2 * phiP + 1 / r_outer * phiDP)

    A = (-3 * EP + r * EDP) / r ** 2
    B = (2 * EP - r * EDP) / r ** 3
    C = -E + 0.5 * r * EP - (1/12) * r ** 2 * EDP
    
    return [A, B, C]


def dRij_dRm_norm(Rij, ijm_list):
    """Calculate the derivative of Rij norm w. r. t. atom m. This term affects
    only on i and j.

    Parameters
    ----------
    Rij : array [j, 3] or [j*k, 3]
        The vector distances of atom i to atom j.
    ijm_list: array [j, 3] or [j*k, 3]
        Id list of center atom i, neighbors atom j, and atom m.

    Returns
    -------
    dRij_m: array [j, 3]
        The derivative of pair atoms w.r.t. atom m in x, y, z directions.
    """
    dRij_m = np.zeros([len(Rij), 3])
    i_dRij_m = np.zeros([len(Rij), 3]) 
    j_dRij_m = np.zeros([len(Rij), 3])
    R1ij = np.linalg.norm(Rij, axis=1).reshape([len(Rij),1])
    
    l1 = (ijm_list[:,2]==ijm_list[:,0])
    dRij_m[l1, :] = -Rij[l1]/R1ij[l1]
    i_dRij_m[l1, :] = -Rij[l1]/R1ij[l1]

    l2 = (ijm_list[:,2]==ijm_list[:,1])
    dRij_m[l2, :] = Rij[l2]/R1ij[l2]
    j_dRij_m[l2, :] = Rij[l2]/R1ij[l2]
    
    l3 = (ijm_list[:,0]==ijm_list[:,1])
    dRij_m[l3, :] = 0

    return dRij_m, i_dRij_m, j_dRij_m


if __name__ == '__main__':
    import time
    from ase.build import bulk
    from ase.io import read
    np.set_printoptions(formatter={'float': '{: 0.6f}'.format})

    t0 = time.time()
    #struc = read("MOD_NiMo.cif")
    #struc = read("MOD_NiMo_real.cif")
    for inner in [1.0, 1.5, 2.0, 2.5, 3.0]:
        print("inner==================", inner)
        for a in [2.0, 3.0, 4.0, 4.5, 5.0, 5.5]:
            struc = bulk('Si', 'diamond', a=a, cubic=True)
            zbl = ZBL(inner, 4.0)
            d = zbl.calculate(struc)
            energy = d['energy'] / len(struc)
            forces = d['force']
            stress = d['stress'] #* 1602176.6208

            #print("Energy: ", energy)
            #print("Force: ")
            #print(forces)
            print(a, "Stress: ", stress)
    t1 = time.time()
    print("\nTime: ", round(t1-t0, 6), "s")
